# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
import FreeCAD
from Path.Tool.machine.models.machine import Machine


class TestPathToolMachine(unittest.TestCase):
    def setUp(self):
        self.default_machine = Machine()

    def test_initialization_defaults(self):
        self.assertEqual(self.default_machine.label, "Machine")
        self.assertAlmostEqual(self.default_machine.max_power.getValueAs("W").Value, 2000)
        self.assertAlmostEqual(self.default_machine.get_min_rpm_value(), 3000)
        self.assertAlmostEqual(self.default_machine.get_max_rpm_value(), 60000)
        self.assertAlmostEqual(self.default_machine.min_feed.getValueAs("mm/min").Value, 1)
        self.assertAlmostEqual(self.default_machine.max_feed.getValueAs("mm/min").Value, 2000)
        expected_peak_torque_rpm = 60000 / 3
        self.assertAlmostEqual(
            self.default_machine.get_peak_torque_rpm_value(),
            expected_peak_torque_rpm,
        )
        expected_max_torque_nm = 2000 * 9.5488 / expected_peak_torque_rpm
        self.assertAlmostEqual(
            self.default_machine.max_torque.getValueAs("Nm").Value,
            expected_max_torque_nm,
        )
        self.assertIsNotNone(self.default_machine.id)

    def test_initialization_custom_values(self):
        custom_machine = Machine(
            label="Custom Machine",
            max_power=5,
            min_rpm=1000,
            max_rpm=20000,
            max_torque=50,
            peak_torque_rpm=15000,
            min_feed=10,
            max_feed=5000,
            id="custom-id",
        )
        self.assertEqual(custom_machine.label, "Custom Machine")
        self.assertAlmostEqual(custom_machine.max_power.getValueAs("W").Value, 5000)
        self.assertAlmostEqual(custom_machine.get_min_rpm_value(), 1000)
        self.assertAlmostEqual(custom_machine.get_max_rpm_value(), 20000)
        self.assertAlmostEqual(custom_machine.max_torque.getValueAs("Nm").Value, 50)
        self.assertAlmostEqual(custom_machine.get_peak_torque_rpm_value(), 15000)
        self.assertAlmostEqual(custom_machine.min_feed.getValueAs("mm/min").Value, 10)
        self.assertAlmostEqual(custom_machine.max_feed.getValueAs("mm/min").Value, 5000)
        self.assertEqual(custom_machine.id, "custom-id")

    def test_initialization_custom_torque_quantity(self):
        custom_torque_machine = Machine(max_torque=FreeCAD.Units.Quantity(100, "Nm"))
        self.assertAlmostEqual(custom_torque_machine.max_torque.getValueAs("Nm").Value, 100)

    def test_validate_valid(self):
        try:
            self.default_machine.validate()
        except AttributeError as e:
            self.fail(f"Validation failed unexpectedly: {e}")

    def test_validate_missing_label(self):
        self.default_machine.label = ""
        with self.assertRaisesRegex(AttributeError, "Machine name is required"):
            self.default_machine.validate()

    def test_validate_peak_torque_rpm_greater_than_max_rpm(self):
        self.default_machine.set_peak_torque_rpm(70000)
        with self.assertRaisesRegex(AttributeError, "Peak Torque RPM.*must be less than max RPM"):
            self.default_machine.validate()

    def test_validate_max_rpm_less_than_min_rpm(self):
        self.default_machine = Machine()
        self.default_machine.set_min_rpm(4000)  # min_rpm = 4000 RPM
        self.default_machine.set_peak_torque_rpm(1000)  # peak_torque_rpm = 1000 RPM
        self.default_machine._max_rpm = 2000 / 60.0  # max_rpm = 2000 RPM (33.33 1/s)
        self.assertLess(
            self.default_machine.get_max_rpm_value(),
            self.default_machine.get_min_rpm_value(),
        )
        with self.assertRaisesRegex(AttributeError, "Max RPM must be larger than min RPM"):
            self.default_machine.validate()

    def test_validate_max_feed_less_than_min_feed(self):
        self.default_machine.set_min_feed(1000)
        self.default_machine._max_feed = 500
        with self.assertRaisesRegex(AttributeError, "Max feed must be larger than min feed"):
            self.default_machine.validate()

    def test_get_torque_at_rpm(self):
        torque_below_peak = self.default_machine.get_torque_at_rpm(10000)
        expected_peak_torque_rpm = 60000 / 3
        expected_max_torque_nm = 2000 * 9.5488 / expected_peak_torque_rpm
        expected_torque_below_peak = expected_max_torque_nm / expected_peak_torque_rpm * 10000
        self.assertAlmostEqual(torque_below_peak, expected_torque_below_peak)

        torque_at_peak = self.default_machine.get_torque_at_rpm(
            self.default_machine.get_peak_torque_rpm_value()
        )
        self.assertAlmostEqual(
            torque_at_peak,
            self.default_machine.max_torque.getValueAs("Nm").Value,
        )

        torque_above_peak = self.default_machine.get_torque_at_rpm(50000)
        expected_torque_above_peak = 2000 * 9.5488 / 50000
        self.assertAlmostEqual(torque_above_peak, expected_torque_above_peak)

    def test_set_label(self):
        self.default_machine.label = "New Label"
        self.assertEqual(self.default_machine.label, "New Label")

    def test_set_max_power(self):
        self.default_machine = Machine()
        self.default_machine.set_max_power(5, "hp")
        self.assertAlmostEqual(
            self.default_machine.max_power.getValueAs("W").Value,
            5 * 745.7,
            places=4,
        )
        with self.assertRaisesRegex(AttributeError, "Max power must be positive"):
            self.default_machine.set_max_power(0)

    def test_set_min_rpm(self):
        self.default_machine = Machine()
        self.default_machine.set_min_rpm(5000)
        self.assertAlmostEqual(self.default_machine.get_min_rpm_value(), 5000)
        with self.assertRaisesRegex(AttributeError, "Min RPM cannot be negative"):
            self.default_machine.set_min_rpm(-100)
        self.default_machine = Machine()
        self.default_machine.set_min_rpm(70000)
        self.assertAlmostEqual(self.default_machine.get_min_rpm_value(), 70000)
        self.assertAlmostEqual(self.default_machine.get_max_rpm_value(), 70001)

    def test_set_max_rpm(self):
        self.default_machine = Machine()
        self.default_machine.set_max_rpm(50000)
        self.assertAlmostEqual(self.default_machine.get_max_rpm_value(), 50000)
        with self.assertRaisesRegex(AttributeError, "Max RPM must be positive"):
            self.default_machine.set_max_rpm(0)
        self.default_machine = Machine()
        self.default_machine.set_max_rpm(2000)
        self.assertAlmostEqual(self.default_machine.get_max_rpm_value(), 2000)
        self.assertAlmostEqual(self.default_machine.get_min_rpm_value(), 1999)
        self.default_machine = Machine()
        self.default_machine.set_max_rpm(0.5)
        self.assertAlmostEqual(self.default_machine.get_max_rpm_value(), 0.5)
        self.assertAlmostEqual(self.default_machine.get_min_rpm_value(), 0)

    def test_set_min_feed(self):
        self.default_machine = Machine()
        self.default_machine.set_min_feed(500, "inch/min")
        self.assertAlmostEqual(
            self.default_machine.min_feed.getValueAs("mm/min").Value,
            500 * 25.4,
            places=4,
        )
        with self.assertRaisesRegex(AttributeError, "Min feed cannot be negative"):
            self.default_machine.set_min_feed(-10)
        self.default_machine = Machine()
        self.default_machine.set_min_feed(3000)
        self.assertAlmostEqual(self.default_machine.min_feed.getValueAs("mm/min").Value, 3000)
        self.assertAlmostEqual(self.default_machine.max_feed.getValueAs("mm/min").Value, 3001)

    def test_set_max_feed(self):
        self.default_machine = Machine()
        self.default_machine.set_max_feed(3000)
        self.assertAlmostEqual(self.default_machine.max_feed.getValueAs("mm/min").Value, 3000)
        with self.assertRaisesRegex(AttributeError, "Max feed must be positive"):
            self.default_machine.set_max_feed(0)
        self.default_machine = Machine()
        self.default_machine.set_min_feed(600)
        self.default_machine.set_max_feed(500)
        self.assertAlmostEqual(self.default_machine.max_feed.getValueAs("mm/min").Value, 500)
        self.assertAlmostEqual(self.default_machine.min_feed.getValueAs("mm/min").Value, 499)
        self.default_machine = Machine()
        self.default_machine.set_max_feed(0.5)
        self.assertAlmostEqual(self.default_machine.max_feed.getValueAs("mm/min").Value, 0.5)
        self.assertAlmostEqual(self.default_machine.min_feed.getValueAs("mm/min").Value, 0)

    def test_set_peak_torque_rpm(self):
        self.default_machine = Machine()
        self.default_machine.set_peak_torque_rpm(40000)
        self.assertAlmostEqual(self.default_machine.get_peak_torque_rpm_value(), 40000)
        with self.assertRaisesRegex(AttributeError, "Peak torque RPM cannot be negative"):
            self.default_machine.set_peak_torque_rpm(-100)

    def test_set_max_torque(self):
        self.default_machine = Machine()
        self.default_machine.set_max_torque(200, "in-lbf")
        self.assertAlmostEqual(
            self.default_machine.max_torque.getValueAs("Nm").Value,
            200 * 0.112985,
            places=4,
        )
        with self.assertRaisesRegex(AttributeError, "Max torque must be positive"):
            self.default_machine.set_max_torque(0)

    def test_dump(self):
        try:
            self.default_machine.dump(False)
        except Exception as e:
            self.fail(f"dump() method failed unexpectedly: {e}")


if __name__ == "__main__":
    unittest.main()



# ***************************************************************************
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint initial pressure document object"
__author__ = "Uwe Stöhr"
__url__ = "https://www.freecad.org"

## @package constraint_initialpressure
#  \ingroup FEM
#  \brief constraint initial pressure object

from . import base_fempythonobject


class ConstraintInitialPressure(base_fempythonobject.BaseFemPythonObject):

    Type = "Fem::ConstraintInitialPressure"

    def __init__(self, obj):
        super().__init__(obj)
        self.add_properties(obj)

    def onDocumentRestored(self, obj):
        self.add_properties(obj)

    def add_properties(self, obj):
        if not hasattr(obj, "Pressure"):
            obj.addProperty("App::PropertyPressure", "Pressure", "Parameter", "Initial Pressure")
            obj.setPropertyStatus("Pressure", "LockDynamic")
            # we initialize 1 bar
            obj.Pressure = "100 kPa"



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Unit tests for the Draft workbench, non-GUI only.

From the terminal, run the following:
FreeCAD -t TestDraft

From within FreeCAD, run the following:
import Test, TestDraft
Test.runTestsFromModule(TestDraft)

For the GUI-only tests see TestDraftGui.
"""

# ===========================================================================
# The unit tests can be run from the operating system terminal, or from
# within FreeCAD itself.
#
# The tests can be run using the full 'FreeCAD' executable
# or the console only 'FreeCADCmd' executable. In the latter case
# some functions cannot be tested as the view providers (visual properties)
# are not available.
#
# ===========================================================================
# In the following, first the command to run the test from the operating
# system terminal is listed, followed by the commands to run the test
# from the Python console within FreeCAD.
#
# ===========================================================================
# Run all Draft tests
# ----
# FreeCAD -t TestDraft
#
# >>> import Test, TestDraft
# >>> Test.runTestsFromModule(TestDraft)
#
# ===========================================================================
# Run tests from a specific module (all classes within this module)
# ----
# FreeCAD -t drafttests.test_creation
#
# >>> import Test, drafttests.test_creation
# >>> Test.runTestsFromModule(drafttests.test_creation)
#
# ===========================================================================
# Run tests from a specific class within a module
# ----
# FreeCAD -t drafttests.test_creation.DraftCreation
#
# >>> import Test, drafttests.test_creation
# >>> Test.runTestsFromClass(drafttests.test_creation.DraftCreation)
#
# ===========================================================================
# Run a specific unit test from a class within a module
# ----
# FreeCAD -t drafttests.test_creation.DraftCreation.test_line
#
# >>> import unittest
# >>> one_test = "drafttests.test_creation.DraftCreation.test_line"
# >>> all_tests = unittest.TestLoader().loadTestsFromName(one_test)
# >>> unittest.TextTestRunner().run(all_tests)

# ===========================================================================
# When the full test is run
#     FreeCAD -t TestDraft
#
# all classes that are found in this file are run.
#
# We import the classes from submodules. These classes contain
# the actual unit tests.
#
# The classes will be run in alphabetical order. So, to force
# a particular order of testing we import them with a name
# that follows a defined alphanumeric sequence.

# Import tests
from drafttests.test_import import DraftImport as DraftTest01

# Objects tests
from drafttests.test_creation import DraftCreation as DraftTest02
from drafttests.test_modification import DraftModification as DraftTest03

# Testing the utils module
from drafttests.test_draftgeomutils import TestDraftGeomUtils as DraftTest04

# Handling of file formats tests
# from drafttests.test_svg import DraftSVG as DraftTest05
from drafttests.test_dxf import DraftDXF as DraftTest06

# from drafttests.test_dwg import DraftDWG as DraftTest07
# from drafttests.test_oca import DraftOCA as DraftTest08
# from drafttests.test_airfoildat import DraftAirfoilDAT as DraftTest09
from drafttests.test_array import DraftArray as DraftTest10

# Use the modules so that code checkers don't complain (flake8)
True if DraftTest01 else False
True if DraftTest02 else False
True if DraftTest03 else False
True if DraftTest04 else False
# True if DraftTest05 else False
True if DraftTest06 else False
# True if DraftTest07 else False
# True if DraftTest08 else False
# True if DraftTest09 else False
True if DraftTest10 else False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Widget for editing a list of properties of a DocumentObject."""

import re
from PySide import QtGui, QtCore
from .property import BasePropertyEditorWidget


def _get_label_text(prop_name):
    """Generate a human-readable label from a property name."""
    # Add space before capital letters (CamelCase splitting)
    s1 = re.sub(r"([A-Z][a-z]+)", r" \1", prop_name)
    # Add space before sequences of capitals (e.g., ID) followed by lowercase
    s2 = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1 \2", s1)
    # Add space before sequences of capitals followed by end of string
    s3 = re.sub(r"([A-Z]+)$", r" \1", s2)
    # Remove leading/trailing spaces and capitalize
    return s3.strip().capitalize()


class DocumentObjectEditorWidget(QtGui.QWidget):
    """
    A widget that displays a user friendly form for editing properties of a
    FreeCAD DocumentObject.
    """

    # Signal emitted when any underlying property value might have changed
    propertyChanged = QtCore.Signal()

    def __init__(self, obj=None, properties_to_show=None, property_suffixes=None, parent=None):
        """
        Initialize the editor widget.

        Args:
            obj (App.DocumentObject, optional): The object to edit. Defaults to None.
            properties_to_show (list[str], optional): List of property names to display.
                                                     Defaults to None (shows nothing).
            property_suffixes (dict[str, str], optional): Dictionary mapping property names
                                                          to suffixes for their labels.
                                                          Defaults to None.
            parent (QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self._obj = obj
        self._properties_to_show = properties_to_show if properties_to_show else []
        self._property_suffixes = property_suffixes if property_suffixes else {}
        self._property_editors = {}  # Store {prop_name: editor_widget}

        self._layout = QtGui.QFormLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setFieldGrowthPolicy(QtGui.QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)

        self._populate_form()

    def _clear_form(self):
        """Remove all rows from the form layout."""
        while self._layout.rowCount() > 0:
            self._layout.removeRow(0)
        self._property_editors.clear()

    def _populate_form(self):
        """Create and add property editors to the form."""
        self._clear_form()
        if not self._obj:
            return

        for prop_name in self._properties_to_show:
            # Only create an editor if the property exists on the object
            if not hasattr(self._obj, prop_name):
                continue

            editor_widget = BasePropertyEditorWidget.for_property(self._obj, prop_name, self)
            label_text = _get_label_text(prop_name)
            suffix = self._property_suffixes.get(prop_name)
            if suffix:
                label_text = f"{label_text} ({suffix}):"
            else:
                label_text = f"{label_text}:"

            label = QtGui.QLabel(label_text)
            self._layout.addRow(label, editor_widget)
            self._property_editors[prop_name] = editor_widget

            # Connect the editor's signal to our own signal
            editor_widget.propertyChanged.connect(self.propertyChanged)

    def setObject(self, obj):
        """Set or change the DocumentObject being edited."""
        if obj != self._obj:
            self._obj = obj
            # Re-populate might be too slow if only object changes,
            # better to just re-attach existing editors.
            # self._populate_form()
            for prop_name, editor in self._property_editors.items():
                editor.attachTo(self._obj, prop_name)

    def setPropertiesToShow(self, properties_to_show, property_suffixes=None):
        """Set or change the list of properties to display."""
        self._properties_to_show = properties_to_show if properties_to_show else []
        self._property_suffixes = property_suffixes if property_suffixes else {}
        self._populate_form()  # Rebuild the form completely

    def updateUI(self):
        """Update all child editor widgets from the object's properties."""
        for editor in self._property_editors.values():
            editor.updateWidget()

    def updateObject(self):
        """Update the object's properties from all child editor widgets."""
        # This might not be strictly necessary if signals are connected,
        # but can be useful for explicit save actions.
        for editor in self._property_editors.values():
            editor.updateProperty()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Op.ThreadMilling as PathThreadMilling
import math

from CAMTests.PathTestUtils import PathTestBase


class TestObject(object):
    def __init__(self, orientation, direction, zTop, zBottom):
        self.ThreadOrientation = orientation
        self.Direction = direction
        self.StartDepth = FreeCAD.Units.Quantity(zTop, FreeCAD.Units.Length)
        self.FinalDepth = FreeCAD.Units.Quantity(zBottom, FreeCAD.Units.Length)


def radii(internal, major, minor, toolDia, toolCrest):
    """test radii function for simple testing"""
    if internal:
        return (minor, major)
    return (major, minor)


class TestPathThreadMilling(PathTestBase):
    """Test thread milling basics."""

    def assertRadii(self, have, want):
        self.assertRoughly(have[0], want[0])
        self.assertRoughly(have[1], want[1])

    def assertList(self, have, want):
        self.assertEqual(len(have), len(want))
        for i in range(len(have)):
            self.assertRoughly(have[i], want[i])

    def assertSetupInternal(self, obj, c, begin, end):
        cmd, zBegin, zEnd = PathThreadMilling.threadSetupInternal(
            obj, obj.StartDepth.Value, obj.FinalDepth.Value
        )
        self.assertEqual(cmd, c)
        self.assertEqual(zBegin, begin)
        self.assertEqual(zEnd, end)

    def assertSetupExternal(self, obj, c, begin, end):
        cmd, zBegin, zEnd = PathThreadMilling.threadSetupExternal(
            obj, obj.StartDepth.Value, obj.FinalDepth.Value
        )
        self.assertEqual(cmd, c)
        self.assertEqual(zBegin, begin)
        self.assertEqual(zEnd, end)

    def test00(self):
        """Verify internal radii."""
        self.assertRadii(PathThreadMilling.threadRadii(True, 20, 18, 2, 0), (8, 9.2))
        self.assertRadii(PathThreadMilling.threadRadii(True, 20, 19, 2, 0), (8.5, 9.1))

    def test01(self):
        """Verify internal radii with tool crest."""
        self.assertRadii(PathThreadMilling.threadRadii(True, 20, 18, 2, 0.1), (8, 9.113397))

    def test10(self):
        """Verify internal thread passes."""
        self.assertList(PathThreadMilling.threadPasses(1, radii, True, 10, 9, 0, 0), [10])
        self.assertList(PathThreadMilling.threadPasses(2, radii, True, 10, 9, 0, 0), [9.707107, 10])
        self.assertList(
            PathThreadMilling.threadPasses(5, radii, True, 10, 9, 0, 0),
            [9.447214, 9.632456, 9.774597, 9.894427, 10],
        )

    def test20(self):
        """Verify external radii."""
        self.assertRadii(PathThreadMilling.threadRadii(False, 20, 18, 2, 0), (11, 9.6))
        self.assertRadii(PathThreadMilling.threadRadii(False, 20, 19, 2, 0), (11, 10.3))

    def test21(self):
        """Verify external radii with tool crest."""
        self.assertRadii(PathThreadMilling.threadRadii(False, 20, 18, 2, 0.1), (11, 9.513397))

    def test30(self):
        """Verify external thread passes."""
        self.assertList(PathThreadMilling.threadPasses(1, radii, False, 10, 9, 0, 0), [9])
        self.assertList(PathThreadMilling.threadPasses(2, radii, False, 10, 9, 0, 0), [9.292893, 9])
        self.assertList(
            PathThreadMilling.threadPasses(5, radii, False, 10, 9, 0, 0),
            [9.552786, 9.367544, 9.225403, 9.105573, 9],
        )

    def test40(self):
        """Verify internal right hand thread setup."""

        hand = PathThreadMilling.RightHand

        self.assertSetupInternal(
            TestObject(hand, PathThreadMilling.DirectionConventional, 1, 0), "G2", 1, 0
        )
        self.assertSetupInternal(
            TestObject(hand, PathThreadMilling.DirectionClimb, 1, 0), "G3", 0, 1
        )

    def test41(self):
        """Verify internal left hand thread setup."""

        hand = PathThreadMilling.LeftHand

        self.assertSetupInternal(
            TestObject(hand, PathThreadMilling.DirectionConventional, 1, 0), "G2", 0, 1
        )
        self.assertSetupInternal(
            TestObject(hand, PathThreadMilling.DirectionClimb, 1, 0), "G3", 1, 0
        )

    def test50(self):
        """Verify exteranl right hand thread setup."""

        hand = PathThreadMilling.RightHand

        self.assertSetupExternal(
            TestObject(hand, PathThreadMilling.DirectionClimb, 1, 0), "G2", 1, 0
        )
        self.assertSetupExternal(
            TestObject(hand, PathThreadMilling.DirectionConventional, 1, 0), "G3", 0, 1
        )

    def test51(self):
        """Verify exteranl left hand thread setup."""

        hand = PathThreadMilling.LeftHand

        self.assertSetupExternal(
            TestObject(hand, PathThreadMilling.DirectionClimb, 1, 0), "G2", 0, 1
        )
        self.assertSetupExternal(
            TestObject(hand, PathThreadMilling.DirectionConventional, 1, 0), "G3", 1, 0
        )



# /***************************************************************************
# *   Copyright (c) 2019 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

import weakref

global_stacks = {}  # dict of TVStacks. key = document name.


class TVStack(object):
    index_LUT = None  # Key = id(tempovis_instance). Value = index in the stack.
    stack = None  # list of weakrefs to TV instances. Using weakrefs, so that TempoVis can self-destruct if forgotten.
    document = None

    _rewind_tv = None

    def __init__(self, document):
        self.document = None
        self.index_LUT = {}
        self.stack = []
        from . import TVObserver  # to start the observer

    def insert(self, tv, index=None):
        if index is None:
            index = len(self.stack)
        idtv = id(tv)
        ref = weakref.ref(tv, (lambda _, idtv=idtv, self=self: self._destruction(idtv)))
        self.stack.insert(index, ref)

        self.rebuild_index(index)

        tv._inserted(self, index)

    def _destruction(self, idtv):
        # the tempovis itself is destroyed already. Purge it from the stack (it should have withdrawn itself, but just in case).
        try:
            index = self.index_LUT.get(idtv)
        except KeyError:
            # already withdrawn
            pass
        else:
            self.stack.pop(index)
            self.index_LUT.pop(idtv)
            self.rebuild_index(index)

    def withdraw(self, tv):
        idtv = id(tv)
        index = self.index_LUT[idtv]
        ref = self.stack.pop(index)
        self.index_LUT.pop(idtv)

        self.rebuild_index(index)

        tv = ref()
        if tv:
            tv._withdrawn(self, index)

    def value_after(self, tv, detail):
        """value_after(tv, detail): returns tuple (tv1, detail), or None.
        Here, tv1 is the tv that remembers the value, and detail is reference to recorded
        data in tv1. None is returned, if no TVs in the stack after the provided one have
        recorded a change to this detail.

        tv can be None, then, the function returns the original value of the detail, or
        None, if the current value matches the original."""
        from . import mTempoVis

        index = self.index_LUT[id(tv)] if tv is not None else -1
        for tvref in self.stack[index + 1 :]:
            tv = tvref()
            if tv.state == mTempoVis.S_ACTIVE:
                if tv.has(detail):
                    return (tv, tv.data[detail.full_key])
        return None

    def rebuild_index(self, start=0):
        if start == 0:
            self.index_LUT = {}
        for i in range(start, len(self.stack)):
            self.index_LUT[id(self.stack[i]())] = i

    def purge_dead(self):
        """removes dead TV instances from the stack"""
        n = 0
        for i in reversed(range(len(self.stack))):
            if self.stack[i]() is None:
                self.stack.pop(i)
                n += 1
        if n > 0:
            self.rebuild_index()
        return n

    def dissolve(self):
        """silently cleans all TVs, so that they won't restore."""
        for ref in self.stack:
            if ref() is not None:
                ref().forget()

    def unwindForSaving(self):
        from . import mTempoVis

        self.rewindAfterSaving()  # just in case there was a failed save before.

        details = (
            {}
        )  # dict of detail original values. Key = detail key; value = detail instance with data representing the original value
        for ref in self.stack:
            tv = ref()
            for key, detail in tv.data.items():
                if not key in details:
                    if detail.affects_persistence:
                        details[detail.full_key] = detail

        self._rewind_tv = mTempoVis.TempoVis(self.document, None)
        for key, detail in details.items():
            self._rewind_tv.modify(detail)

    def rewindAfterSaving(self):
        if self._rewind_tv is not None:
            self._rewind_tv.restore()
            self._rewind_tv = None

    def getSplitSequence(self, tv):
        """getSplitSequence(tv): returns (list_before, list_after), neither list includes tv."""
        index = self.index_LUT[id(tv)]

        def deref(lst):
            return [ref() for ref in lst]

        return deref(self.stack[0:index]), deref(self.stack[index + 1 :])

    def __getitem__(self, index):
        return self.stack[index]()

    def __len__(self):
        return len(self.stack)

    def __iter__(self):
        for ref in self.stack:
            yield ref()

    def __reversed__(self):
        for ref in reversed(self.stack):
            yield ref()

    def restoreAll(self):
        for ref in reversed(self.stack):
            ref().restore()

    def byTag(self, tag):
        return [ref() for ref in self.stack if ref().tag == tag]


def mainStack(document, create_if_missing=True):
    """mainStack(document, create_if_missing = True):returns the main TVStack instance for provided document"""
    docname = document.Name

    if create_if_missing:
        if not docname in global_stacks:
            global_stacks[docname] = TVStack(document)

    return global_stacks.get(docname, None)


def _slotDeletedDocument(document):
    docname = document.Name
    stk = global_stacks.pop(docname, None)
    if stk is not None:
        stk.dissolve()


def _slotStartSaveDocument(doc):
    stk = mainStack(doc, create_if_missing=False)
    if stk is not None:
        stk.unwindForSaving()


def _slotFinishSaveDocument(doc):
    stk = mainStack(doc, create_if_missing=False)
    if stk is not None:
        stk.rewindAfterSaving()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *   Copyright (c) 2025 Billy Huddleston <billy@ivdc.com>                  *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtGui, QtCore
import FreeCAD
import FreeCADGui
import Path.Log
import os
import tempfile


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class ImageBuilder:
    def __init__(self, file_path):
        self.file_path = file_path

    def build_image(self, obj, image_name, as_bytes=False, view="default"):
        raise NotImplementedError("Subclass must implement abstract method")

    def save_image(self, image):
        raise NotImplementedError("Subclass must implement abstract method")


class ImageBuilderFactory:
    @staticmethod
    def get_image_builder(file_path, **kwargs):
        # return DummyImageBuilder(file_path, **kwargs)
        if FreeCAD.GuiUp:
            return GuiImageBuilder(file_path, **kwargs)
        else:
            return DummyImageBuilder(file_path, **kwargs)
        # return NonGuiImageBuilder(file_path, **kwargs)


class DummyImageBuilder(ImageBuilder):
    def __init__(self, file_path):
        Path.Log.debug("Initializing dummyimagebuilder")
        super().__init__(file_path)

    def build_image(self, obj, imageName, as_bytes=False, view="default"):
        if as_bytes:
            return b""
        return self.file_path


class GuiImageBuilder(ImageBuilder):
    """
    A class for generating images of 3D objects in a FreeCAD GUI environment.
    """

    def __init__(self, file_path):
        super().__init__(file_path)

        Path.Log.debug("Initializing GuiImageBuilder")
        self.file_path = file_path
        self.currentCamera = FreeCADGui.ActiveDocument.ActiveView.getCameraType()

        self.doc = FreeCADGui.ActiveDocument

    def __del__(self):
        Path.Log.debug("Destroying GuiImageBuilder")
        self.restore_visibility()

    def prepare_view(self, obj, view="default"):
        # Create a new view
        Path.Log.debug("CAM - Preparing view\n")

        mw = FreeCADGui.getMainWindow()
        num_windows = len(mw.getWindows())

        # Create and configure the view
        view_obj = FreeCADGui.ActiveDocument.createView("Gui::View3DInventor")
        view_obj.setAnimationEnabled(False)
        view_obj.setCameraType("Orthographic")
        if view == "headon":
            view_obj.viewFront()
        else:
            view_obj.viewIsometric()

        # Resize the window
        mdi = mw.findChild(QtGui.QMdiArea)
        view_window = mdi.activeSubWindow()
        view_window.resize(500, 500)
        view_window.showMaximized()

        # First make everything invisible
        self.record_visibility()

        # Then make only our target object visible and select it
        obj.Visibility = True
        FreeCADGui.Selection.clearSelection()
        FreeCADGui.Selection.addSelection(obj)
        FreeCADGui.Selection.clearSelection()  # Clear so the selection highlight does not appear in the image

        # Get the active view and fit to selection
        a_view = FreeCADGui.activeDocument().activeView()
        try:
            a_view.fitAll()  # First fit all to ensure the object is in view
            FreeCADGui.updateGui()
            a_view.fitSelection()  # Then try to fit to the selection
        except Exception:
            # If fitSelection fails, we already called fitAll
            pass

        # Return the index of the new window (= old number of windows)
        return num_windows

    def record_visibility(self):
        self.visible = [o for o in self.doc.Document.Objects if o.Visibility]
        for o in self.doc.Document.Objects:
            o.Visibility = False

    def destroy_view(self, idx):
        Path.Log.debug("CAM - destroying view\n")
        mw = FreeCADGui.getMainWindow()
        windows = mw.getWindows()
        mw.removeWindow(windows[idx])

    def restore_visibility(self):
        Path.Log.debug("CAM - Restoring visibility\n")
        for o in self.visible:
            o.Visibility = True

    def build_image(self, obj, image_name, as_bytes=False, view="default"):
        Path.Log.debug("CAM - Building image\n")
        """
        Makes an image of the target object. Returns either the image as bytes or a filename.
        """

        idx = self.prepare_view(obj, view=view)

        if as_bytes:
            # Capture directly to memory without writing to disk
            img_bytes = self.capture_image_to_bytes()
            self.destroy_view(idx)
            return img_bytes
        else:
            # Write to disk as before
            file_path = os.path.join(self.file_path, image_name)
            self.capture_image(file_path)
            self.destroy_view(idx)
            result = f"{file_path}_t.png"
            Path.Log.debug(f"Image saved to: {result}")
            return result

    def capture_image(self, file_path):
        FreeCADGui.updateGui()
        Path.Log.debug("CAM - capture image to file\n")
        a_view = FreeCADGui.activeDocument().activeView()
        # Generate higher resolution images - 800x800 pixels for better quality on high-DPI displays
        a_view.saveImage(file_path + ".png", 800, 800, "Current")
        a_view.saveImage(file_path + "_t.png", 800, 800, "Transparent")
        a_view.setAnimationEnabled(True)

    def capture_image_to_bytes(self):
        """Capture the current view directly to bytes without writing to disk"""
        FreeCADGui.updateGui()
        Path.Log.debug("CAM - capture image to bytes\n")
        a_view = FreeCADGui.activeDocument().activeView()

        try:
            # Use FreeCAD's built-in method for getting a QImage directly
            # This approach is based on the same method the viewport uses internally
            qimg = a_view.grabFramebuffer()

            # Convert QImage to bytes using QBuffer (purely in memory)
            buffer = QtCore.QBuffer()
            buffer.open(QtCore.QIODevice.WriteOnly)
            qimg.save(buffer, "PNG")
            img_bytes = buffer.data().data()
            buffer.close()

            a_view.setAnimationEnabled(True)
            return img_bytes

        except Exception as e:
            # Fallback to temporary file approach if the direct method fails
            Path.Log.debug(f"Direct image capture failed: {e}, using fallback method")

            with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as temp:
                temp_path = temp.name

            # saveImage doesn't write to memory directly, so we need to use a temporary file
            # Generate higher resolution images - 800x800 pixels for better quality on high-DPI displays
            a_view.saveImage(temp_path, 800, 800, "Transparent")

            # Read the temporary file into memory
            with open(temp_path, "rb") as f:
                img_bytes = f.read()

            # Clean up the temporary file
            try:
                os.unlink(temp_path)
            except Exception:
                # Ignore errors during temporary file cleanup, as failure to delete is non-critical
                pass

            a_view.setAnimationEnabled(True)
            return img_bytes


class NonGuiImageBuilder(ImageBuilder):
    def __init__(self, file_path):
        super().__init__(file_path)
        Path.Log.debug("nonguiimagebuilder")

    def build_image(self, obj, image_name, as_bytes=False, view="default"):
        """
        Generates a headless picture of a 3D object and saves it as a PNG and optionally a PostScript file.

        Args:
        - obj: The 3D object to generate an image for.
        - image_name: Base name for the output image file without extension.

        Returns:
        - A boolean indicating the success of the operation.
        """
        # Ensure the 'Part' and 'coin' modules are available, along with necessary attributes/methods.
        if not hasattr(obj, "Shape") or not hasattr(obj.Shape, "writeInventor"):
            Path.Log.debug("Object does not have the required attributes.")
            return False

        try:
            # Generate Inventor data from the object's shape
            iv = obj.Shape.writeInventor()

            # Prepare Inventor data for rendering
            inp = coin.SoInput()
            inp.setBuffer(iv)
            data = coin.SoDB.readAll(inp)

            if data is None:
                Path.Log.debug("Failed to read Inventor data.")
                return False

            # Setup the scene
            base = coin.SoBaseColor()
            base.rgb.setValue(0.6, 0.7, 1.0)
            data.insertChild(base, 0)

            root = coin.SoSeparator()
            light = coin.SoDirectionalLight()
            cam = coin.SoOrthographicCamera()
            root.addChild(cam)
            root.addChild(light)
            root.addChild(data)

            # Camera and rendering setup
            axo = coin.SbRotation(-0.353553, -0.146447, -0.353553, -0.853553)
            viewport = coin.SbViewportRegion(400, 400)
            cam.orientation.setValue(axo)
            cam.viewAll(root, viewport)
            off = coin.SoOffscreenRenderer(viewport)
            root.ref()
            ret = off.render(root)
            root.unref()
            if as_bytes:
                qimg = off.getQImage()
                buffer = QtCore.QBuffer()
                buffer.open(QtCore.QIODevice.WriteOnly)
                qimg.save(buffer, "PNG")
                return buffer.data().data()
            else:
                if off.isWriteSupported("PNG"):
                    file_path = f"{self.file_path}{os.path.sep}{image_name}.png"
                    off.writeToFile(file_path, "PNG")
                    return file_path
                else:
                    Path.Log.debug("PNG format is not supported.")
                    return False

        except Exception as e:
            print(f"An error occurred: {e}")
            return False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 imarin                                             *
# *                                                                         *
# *   Heavily based on gbrl post-procesor by:                               *
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import datetime
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
from builtins import open as pyopen

now = datetime.datetime.now()

"""
Generate g-code compatible with fablin from a Path.

import fablin_post
fablin_post.export(object,"/path/to/file.ncc")
"""

TOOLTIP_ARGS = """
Arguments for fablin:
    --rapids-feedrate                ... feedrate to be used for rapids (e.g. --rapids-feedrate=300)
    --header,--no-header             ... output headers (--header)
    --comments,--no-comments         ... output comments (--comments)
    --line-numbers,--no-line-numbers ... prefix with line numbers (--no-lin-numbers)
    --show-editor, --no-show-editor  ... pop up editor before writing output(--show-editor)
"""

# These globals set common customization preferences
OUTPUT_COMMENTS = False  # Fablin does not support parenthesis, it will echo the command complaining. As a side effect the spinner may turn at a very reduced speed (do not ask me why).
OUTPUT_HEADER = False  # Same as above. You can enable this by passing arguments to the post-processor in case you need them for example for debugging the gcode.
OUTPUT_LINE_NUMBERS = False
OUTPUT_TOOL_CHANGE = False

SHOW_EDITOR = True
MODAL = False  # if true commands are suppressed if the same as previous line.
COMMAND_SPACE = " "
LINENR = 100  # line number starting value

# These globals will be reflected in the Machine configuration of the project
UNITS = ""  # only metric, G20/G21 is ignored
MACHINE_NAME = "FABLIN"
CORNER_MIN = {"x": 0, "y": 0, "z": 0}
CORNER_MAX = {"x": 500, "y": 300, "z": 300}

RAPID_MOVES = ["G0", "G00"]

RAPID_FEEDRATE = 10000

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = """G90
"""

# Postamble text will appear following the last operation.
POSTAMBLE = """M5
G00 X-1.0 Y1.0
G90
"""

# These commands are ignored by commenting them out
SUPPRESS_COMMANDS = ["G98", "G80", "M6", "G17"]

# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = """"""


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global RAPID_FEEDRATE
    for arg in argstring.split():
        if arg == "--header":
            OUTPUT_HEADER = True
        elif arg == "--no-header":
            OUTPUT_HEADER = False
        elif arg == "--comments":
            OUTPUT_COMMENTS = True
        elif arg == "--no-comments":
            OUTPUT_COMMENTS = False
        elif arg == "--line-numbers":
            OUTPUT_LINE_NUMBERS = True
        elif arg == "--no-line-numbers":
            OUTPUT_LINE_NUMBERS = False
        elif arg == "--show-editor":
            SHOW_EDITOR = True
        elif arg == "--no-show-editor":
            SHOW_EDITOR = False

        params = arg.split("=")

        if params[0] == "--rapids-feedrate":
            RAPID_FEEDRATE = params[1]


def export(objectslist, filename, argstring):
    processArguments(argstring)
    global UNITS
    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return

    print("postprocessing...")
    gcode = ""

    # Find the machine.
    # The user my have overridden post processor defaults in the GUI.
    # Make sure we're using the current values in the Machine Def.
    myMachine = None
    for pathobj in objectslist:
        if hasattr(pathobj, "Group"):  # We have a compound or project.
            for p in pathobj.Group:
                if p.Name == "Machine":
                    myMachine = p
    if myMachine is None:
        print("No machine found in this project")
    else:
        if myMachine.MachineUnits == "Metric":
            UNITS = "G21"
        else:
            UNITS = "G20"

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(begin preamble)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    for obj in objectslist:

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(begin operation: " + obj.Label + ")\n"
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(finish operation: " + obj.Label + ")\n"
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

    # do the post_amble

    if OUTPUT_COMMENTS:
        gcode += "(begin postamble)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    if SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    print("done postprocessing.")

    if not filename == "-":
        gfile = pyopen(filename, "w")
        gfile.write(final)
        gfile.close()

    return final


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 10
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    out = ""
    lastcommand = None

    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "I",
        "J",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
    ]  # linuxcnc doesn't want K properties on XY plane  Arcs need work.

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        if OUTPUT_COMMENTS:
            out += linenumber() + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        if not hasattr(pathobj, "Path"):  # groups might contain non-path things like stock.
            return out

        if OUTPUT_COMMENTS:
            out += linenumber() + "(Path: " + pathobj.Label + ")\n"

        for c in PathUtils.getPathWithPlacement(pathobj).Commands:
            outstring = []
            command = c.Name

            # fablin does not support parenthesis syntax, so removing that (pocket) in the agnostic gcode
            if command.startswith("("):
                if not OUTPUT_COMMENTS:
                    pass
            else:
                outstring.append(command)

            # if modal: only print the command if it is not the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F":
                        if command not in RAPID_MOVES:
                            outstring.append(param + format(c.Parameters["F"], ".2f"))
                    elif param == "T":
                        outstring.append(param + str(c.Parameters["T"]))
                    else:
                        outstring.append(param + format(c.Parameters[param], ".4f"))

            if command in RAPID_MOVES and command != lastcommand:
                outstring.append("F" + format(RAPID_FEEDRATE))

            # store the latest command
            lastcommand = command

            # Check for Tool Change:
            if command == "M6":
                if OUTPUT_COMMENTS:
                    out += linenumber() + "(begin toolchange)\n"
                if not OUTPUT_TOOL_CHANGE:
                    outstring.insert(0, ";")
                else:
                    for line in TOOL_CHANGE.splitlines(True):
                        out += linenumber() + line

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            if command in SUPPRESS_COMMANDS:
                outstring = []

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                for w in outstring:
                    out += w + COMMAND_SPACE
                out = out.strip() + "\n"

        return out


# print(__name__ + " gcode postprocessor loaded.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# This package aggregates tool bit shape classes.

# Import the base class and all concrete shape classes
from .models.base import ToolBitShape
from .models.ballend import ToolBitShapeBallend
from .models.bullnose import ToolBitShapeBullnose
from .models.chamfer import ToolBitShapeChamfer
from .models.custom import ToolBitShapeCustom
from .models.dovetail import ToolBitShapeDovetail
from .models.drill import ToolBitShapeDrill
from .models.endmill import ToolBitShapeEndmill
from .models.radius import ToolBitShapeRadius
from .models.probe import ToolBitShapeProbe
from .models.reamer import ToolBitShapeReamer
from .models.slittingsaw import ToolBitShapeSlittingSaw
from .models.tap import ToolBitShapeTap
from .models.taperedballnose import ToolBitShapeTaperedBallNose
from .models.threadmill import ToolBitShapeThreadMill
from .models.vbit import ToolBitShapeVBit
from .models.icon import (
    ToolBitShapeIcon,
    ToolBitShapePngIcon,
    ToolBitShapeSvgIcon,
)

# A list of the name of each ToolBitShape
TOOL_BIT_SHAPE_NAMES = sorted([cls.name for cls in ToolBitShape.__subclasses__()])

# Define __all__ for explicit public interface
__all__ = [
    "ToolBitShape",
    "ToolBitShapeBallend",
    "ToolBitShapeBullnose",
    "ToolBitShapeChamfer",
    "ToolBitShapeCustom",
    "ToolBitShapeDovetail",
    "ToolBitShapeDrill",
    "ToolBitShapeEndmill",
    "ToolBitShapeRadius",
    "ToolBitShapeProbe",
    "ToolBitShapeReamer",
    "ToolBitShapeSlittingSaw",
    "ToolBitShapeTap",
    "ToolBitShapeTaperedBallNose",
    "ToolBitShapeThreadMill",
    "ToolBitShapeVBit",
    "TOOL_BIT_SHAPE_NAMES",
    "ToolBitShapeIcon",
    "ToolBitShapeSvgIcon",
    "ToolBitShapePngIcon",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from Path.Post.Processor import PostProcessor
import Path
import FreeCAD
import math

Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate

debug = False
if debug:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

# Define colors for the layers
LAYER_COLORS = {"CUT": "red", "ENGRAVE": "blue", "FILL": "green", "DEFAULT": "black"}


class Svg(PostProcessor):
    def __init__(self, job):

        super().__init__(
            job,
            tooltip=translate("CAM", "SVG post processor"),
            tooltipargs=[],
            units="mm",
        )
        Path.Log.debug("SVG post processor initialized")

    def export(self):
        Path.Log.debug("Exporting the job")

        use_layers = "--layers" in self._job.PostProcessorArgs

        postables = self._buildPostList()
        Path.Log.debug(f"postables count: {len(postables)}")

        svg_strings = []
        for idx, section in enumerate(postables):
            svg_content = self.create_svg_section(section, idx, use_layers)
            svg_strings.append((idx, svg_content))

        return svg_strings

    def create_svg_section(self, section, idx, use_layers):
        Path.Log.track()
        partname, sublist = section

        # Initialize bounding box
        xmin, ymin, xmax, ymax = self.calculate_bounding_box(sublist)

        if xmin is None or ymin is None or xmax is None or ymax is None:
            Path.Log.debug("No wires found, skipping section")
            return ""

        width = xmax - xmin
        height = ymax - ymin

        # Create the SVG header with a normalized viewBox
        svg_content = (
            f'<svg xmlns="http://www.w3.org/2000/svg" '
            f'xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape" '
            f'width="{width}mm" height="{height}mm" viewBox="0 0 {width} {height}">\n'
        )

        # Convert each wire to an SVG path
        for obj_idx, obj in enumerate(sublist):

            strokestyle = self.get_stroke_style(obj)
            color = strokestyle["color"]
            width = strokestyle["width"]
            pathtype = strokestyle["label"]
            Path.Log.debug(pathtype)

            wires = Path.Geom.wiresForPath(obj.Path)
            if not wires:
                continue  # Skip objects that do not produce wires

            if use_layers:
                layer_id = f"layer{idx}_{obj_idx}"
                svg_content += (
                    f'<g id="{layer_id}" inkscape:label="{obj.Label}" inkscape:groupmode="layer">\n'
                )

            for wire in wires:
                path_data = self.wire_to_svg_path(wire, width, height, xmin, ymin)
                if pathtype == "FILL":
                    svg_content += f'  <path d="{path_data}" stroke="none" fill="{color}" />\n'
                else:
                    svg_content += f'  <path d="{path_data}" stroke="{color}" stroke-width="{width}" fill="none" />\n'

            if use_layers:
                svg_content += "</g>\n"

        # Close the SVG tag
        svg_content += "</svg>"

        return svg_content

    def wire_to_svg_path(self, wire, width, height, xmin, ymin):
        """Convert FreeCAD wire to SVG path data with y-axis inversion and limited precision"""
        path_data = ""
        is_first_point = True
        vertices_info = []

        def format_coord(value):
            return f"{value:.4f}"

        for edge in wire.Edges:
            start_point = edge.Vertexes[0].Point
            end_point = edge.Vertexes[-1].Point
            Path.Log.debug(f"Edge Type: {edge.Curve.TypeId}")
            Path.Log.debug(
                f"Start Point: ({format_coord(start_point.x)}, {format_coord(start_point.y)})"
            )
            Path.Log.debug(f"End Point: ({format_coord(end_point.x)}, {format_coord(end_point.y)})")

            # Check if the edge is vertical (should be skipped)
            if start_point.x == end_point.x and start_point.y == end_point.y:
                Path.Log.debug("Skipping vertical edge")
                continue

            if is_first_point:
                path_data += f"M {format_coord(start_point.x - xmin)} {format_coord(height - (start_point.y - ymin))} "
                vertices_info.append(
                    f"M {format_coord(start_point.x - xmin)} {format_coord(height - (start_point.y - ymin))}"
                )
                is_first_point = False

            if edge.Curve.TypeId in ["Part::GeomLineSegment", "Part::GeomLine"]:
                # Handle line segment without discretization
                path_data += f"L {format_coord(end_point.x - xmin)} {format_coord(height - (end_point.y - ymin))} "
                vertices_info.append(
                    f"L {format_coord(end_point.x - xmin)} {format_coord(height - (end_point.y - ymin))}"
                )
                Path.Log.debug(
                    f"Line segment from ({format_coord(start_point.x)}, {format_coord(start_point.y)}) to ({format_coord(end_point.x)}, {format_coord(end_point.y)})"
                )
            elif edge.Curve.TypeId in ["Part::GeomCircle", "Part::GeomArcOfCircle"]:
                # Handle circular arc using 'A' command
                radius = edge.Curve.Radius
                center = edge.Curve.Center
                start_angle = math.atan2(start_point.y - center.y, start_point.x - center.x)
                end_angle = math.atan2(end_point.y - center.y, end_point.x - center.x)

                # Calculate the angle difference and normalize to [-π, π]
                angle_diff = (end_angle - start_angle) % (2 * math.pi)
                if angle_diff > math.pi:
                    angle_diff -= 2 * math.pi

                Path.Log.debug(f"Angle difference: {angle_diff:.2f} radians")

                # Determine the large_arc_flag and sweep_flag
                large_arc_flag = 1 if abs(angle_diff) > (math.pi) else 0
                sweep_flag = 0 if angle_diff > 0 else 1

                start_x = format_coord(start_point.x - xmin)
                start_y = format_coord(height - (start_point.y - ymin))
                end_x = format_coord(end_point.x - xmin)
                end_y = format_coord(height - (end_point.y - ymin))

                path_data += f"A {format_coord(radius)} {format_coord(radius)} 0 {large_arc_flag} {sweep_flag} {end_x} {end_y} "
                vertices_info.append(
                    f"A {format_coord(radius)} {format_coord(radius)} 0 {large_arc_flag} {sweep_flag} {end_x} {end_y}"
                )
                Path.Log.debug(
                    f"Circular arc with radius {format_coord(radius)} from ({start_x}, {start_y}) to ({end_x}, {end_y}) (large_arc_flag: {large_arc_flag}, sweep_flag: {sweep_flag})"
                )
                Path.Log.debug(path_data)
            else:
                # Discretize other types of edges into 100 segments
                vertices = edge.discretize(100)
                if len(vertices) < 2:
                    continue
                if all(vertices[0].x == v.x and vertices[0].y == v.y for v in vertices[1:]):
                    continue
                for vertex in vertices:
                    path_data += f"L {format_coord(vertex.x - xmin)} {format_coord(height - (vertex.y - ymin))} "
                    vertices_info.append(
                        f"L {format_coord(vertex.x - xmin)} {format_coord(height - (vertex.y - ymin))}"
                    )
                Path.Log.debug(f"Discretized edge with {len(vertices)} points")

        path_data += "Z "  # Close the path for filled edges
        vertices_info.append("Z")
        Path.Log.debug(f"SVG Path: {path_data.strip()}")
        Path.Log.debug(f"Vertices Info: {vertices_info}")
        return path_data.strip()

    def calculate_bounding_box(self, sublist):
        xmin, ymin, xmax, ymax = None, None, None, None
        for obj in sublist:
            wires = Path.Geom.wiresForPath(obj.Path)
            for wire in wires:
                for vertex in wire.Vertexes:
                    x, y = vertex.X, vertex.Y
                    if xmin is None or x < xmin:
                        xmin = x
                    if xmax is None or x > xmax:
                        xmax = x
                    if ymin is None or y < ymin:
                        ymin = y
                    if ymax is None or y > ymax:
                        ymax = y
        return xmin, ymin, xmax, ymax

    def get_stroke_style(self, obj):
        strokewidth = 0.1
        color = LAYER_COLORS["DEFAULT"]
        label = None

        if hasattr(obj, "ToolController"):
            tc = obj.ToolController
            if hasattr(tc, "Tool") and hasattr(tc.Tool, "Diameter"):
                strokewidth = tc.Tool.Diameter
            if hasattr(tc, "Label"):
                for key in LAYER_COLORS:
                    if key in tc.Label:
                        label = key
                        color = LAYER_COLORS[key]
                        break

        return {"width": strokewidth, "color": color, "label": label}

    @property
    def tooltip(self):
        tooltip = """
        This is the Ondsel SVG CAM post processor.
        It will export a CAM job to an SVG file with colors and layers.

        Tool Controllers determine how the svg will be created and, thus, how the
        laser will behave.  They must have a label containing a string from the list at the top of the post file.
        (e.g. CUT, FILL, ENGRAVE). SVG Paths will be color coded similarly

        Actual laser behavior will depend on how the laser controller is configured
        to process the colors / layers.

        Step-downs in operations will result in multiple passes.


        """
        return tooltip

    @property
    def tooltipArgs(self):
        argtooltip = """
        --layers: Output will be written to different layers. Layer names are taken from the operation label
        """
        return argtooltip

    @property
    def preferredExtension(self):
        return "svg"



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Open files FEM Gui unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import tempfile
import unittest
from os.path import join
from os import remove

import FreeCAD

from femtest.app import support_utils as testtools
from femtest.app.support_utils import fcc_print
from femtest.app.test_object import create_all_fem_objects_doc


"""
FIXME TODO HACK
Important note!
Delete build directory (at least in fem the objects and vpobjects directories)
if not migrate will not be used because the old modules might still be in the
build directory, thus the test will fail
rm -rf Mod/Fem/
FIXME TODO HACK
"""


"""
# TODO: separate unit test:
# std document name of object ==  obj type
ATM:
for elmer equation obj: name != proxy type
material solid and material fluid obj: name != proxy type

# in addition for FeaturePythons
# std document name of object ==  the class name (for for femsolver obj)
# all femsolver objects class name is Proxy
"""


class TestObjectOpen(unittest.TestCase):
    fcc_print("import TestObjectOpen")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test
        doc_name = self.__class__.__name__
        self.document = FreeCAD.newDocument(doc_name)

        self.test_file_dir = join(testtools.get_fem_test_home_dir(), "open")

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars

        fcc_print(
            "\n{0}\n{1} run FEM TestObjectOpen tests {2}\n{0}".format(100 * "*", 10 * "*", 60 * "*")
        )

    # ********************************************************************************************
    def test_femobjects_open_head(self):
        fcc_print("load master head document objects")

        # get a document with all FEM objects
        self.document = create_all_fem_objects_doc(self.document)

        # save and load the document
        file_path = join(tempfile.gettempdir(), "all_objects_head.FCStd")
        self.document.saveAs(file_path)
        FreeCAD.closeDocument(self.document.Name)
        self.document = FreeCAD.open(file_path)

        # FeaturePythons view provider
        self.compare_feature_pythons_class_gui(self.document)

        # standard name changed
        from femsolver.elmer.equations.flux import ViewProxy

        remove(file_path)
        self.assertEqual(ViewProxy, self.document.Flux.ViewObject.Proxy.__class__)

    # ********************************************************************************************
    def test_femobjects_open_de9b3fb438(self):
        # the number in method name is the FreeCAD commit the document was created with
        # https://github.com/FreeCAD/FreeCAD/commit/de9b3fb438
        # the document was created by running the object create unit test
        # FreeCAD --run-test "femtest.app.test_object.TestObjectCreate.test_femobjects_make"
        fcc_print("load old document objects")
        FreeCAD.closeDocument(self.document.Name)  # close the empty document from setUp first
        self.document = FreeCAD.open(join(self.test_file_dir, "all_objects_de9b3fb438.FCStd"))

        # FeaturePythons view provider
        self.compare_feature_pythons_class_gui(self.document)

        # standard name changed
        from femsolver.elmer.equations.flux import ViewProxy

        self.assertEqual(ViewProxy, self.document.Fluxsolver.ViewObject.Proxy.__class__)

    # ********************************************************************************************
    def compare_feature_pythons_class_gui(self, doc):
        import ObjectsFem
        from femtools.femutils import type_of_obj

        # see comments at file end, the code was created by some python code

        from femviewprovider.view_constraint_bodyheatsource import VPConstraintBodyHeatSource

        self.assertEqual(
            VPConstraintBodyHeatSource, doc.ConstraintBodyHeatSource.ViewObject.Proxy.__class__
        )

        self.assertEqual(
            "Fem::ConstraintCurrentDensity",
            type_of_obj(ObjectsFem.makeConstraintCurrentDensity(doc)),
        )

        from femviewprovider.view_constraint_electrostaticpotential import (
            VPConstraintElectroStaticPotential,
        )

        self.assertEqual(
            VPConstraintElectroStaticPotential,
            doc.ConstraintElectrostaticPotential.ViewObject.Proxy.__class__,
        )

        from femviewprovider.view_constraint_flowvelocity import VPConstraintFlowVelocity

        self.assertEqual(
            VPConstraintFlowVelocity, doc.ConstraintFlowVelocity.ViewObject.Proxy.__class__
        )

        from femviewprovider.view_constraint_initialflowvelocity import (
            VPConstraintInitialFlowVelocity,
        )

        self.assertEqual(
            VPConstraintInitialFlowVelocity,
            doc.ConstraintInitialFlowVelocity.ViewObject.Proxy.__class__,
        )

        self.assertEqual(
            "Fem::ConstraintMagnetization", type_of_obj(ObjectsFem.makeConstraintMagnetization(doc))
        )

        from femviewprovider.view_constraint_selfweight import VPConstraintSelfWeight

        self.assertEqual(
            VPConstraintSelfWeight, doc.ConstraintSelfWeight.ViewObject.Proxy.__class__
        )

        from femviewprovider.view_constraint_tie import VPConstraintTie

        self.assertEqual(VPConstraintTie, doc.ConstraintTie.ViewObject.Proxy.__class__)

        from femviewprovider.view_element_fluid1D import VPElementFluid1D

        self.assertEqual(VPElementFluid1D, doc.ElementFluid1D.ViewObject.Proxy.__class__)

        from femviewprovider.view_element_geometry1D import VPElementGeometry1D

        self.assertEqual(VPElementGeometry1D, doc.ElementGeometry1D.ViewObject.Proxy.__class__)

        from femviewprovider.view_element_geometry2D import VPElementGeometry2D

        self.assertEqual(VPElementGeometry2D, doc.ElementGeometry2D.ViewObject.Proxy.__class__)

        from femviewprovider.view_element_rotation1D import VPElementRotation1D

        self.assertEqual(VPElementRotation1D, doc.ElementRotation1D.ViewObject.Proxy.__class__)

        from femviewprovider.view_material_common import VPMaterialCommon

        self.assertEqual(VPMaterialCommon, doc.MaterialFluid.ViewObject.Proxy.__class__)

        from femviewprovider.view_material_common import VPMaterialCommon

        self.assertEqual(VPMaterialCommon, doc.MaterialSolid.ViewObject.Proxy.__class__)

        from femviewprovider.view_material_mechanicalnonlinear import VPMaterialMechanicalNonlinear

        self.assertEqual(
            VPMaterialMechanicalNonlinear,
            doc.MaterialMechanicalNonlinear.ViewObject.Proxy.__class__,
        )

        from femviewprovider.view_material_reinforced import VPMaterialReinforced

        self.assertEqual(VPMaterialReinforced, doc.MaterialReinforced.ViewObject.Proxy.__class__)

        from femviewprovider.view_mesh_gmsh import VPMeshGmsh

        self.assertEqual(VPMeshGmsh, doc.MeshGmsh.ViewObject.Proxy.__class__)

        from femviewprovider.view_mesh_boundarylayer import VPMeshBoundaryLayer

        self.assertEqual(VPMeshBoundaryLayer, doc.MeshBoundaryLayer.ViewObject.Proxy.__class__)

        from femviewprovider.view_mesh_group import VPMeshGroup

        self.assertEqual(VPMeshGroup, doc.MeshGroup.ViewObject.Proxy.__class__)

        from femviewprovider.view_mesh_region import VPMeshRegion

        self.assertEqual(VPMeshRegion, doc.MeshRegion.ViewObject.Proxy.__class__)

        from femviewprovider.view_mesh_result import VPFemMeshResult

        self.assertEqual(VPFemMeshResult, doc.MeshResult.ViewObject.Proxy.__class__)

        from femviewprovider.view_result_mechanical import VPResultMechanical

        self.assertEqual(VPResultMechanical, doc.ResultMechanical.ViewObject.Proxy.__class__)

        from femviewprovider.view_solver_ccxtools import VPSolverCcxTools

        self.assertEqual(VPSolverCcxTools, doc.SolverCcxTools.ViewObject.Proxy.__class__)

        from femsolver.elmer.solver import ViewProxy

        self.assertEqual(ViewProxy, doc.SolverElmer.ViewObject.Proxy.__class__)

        from femsolver.z88.solver import ViewProxy

        self.assertEqual(ViewProxy, doc.SolverZ88.ViewObject.Proxy.__class__)

        from femsolver.elmer.equations.elasticity import ViewProxy

        self.assertEqual(ViewProxy, doc.Elasticity.ViewObject.Proxy.__class__)

        from femsolver.elmer.equations.electrostatic import ViewProxy

        self.assertEqual(ViewProxy, doc.Electrostatic.ViewObject.Proxy.__class__)

        from femsolver.elmer.equations.flow import ViewProxy

        self.assertEqual(ViewProxy, doc.Flow.ViewObject.Proxy.__class__)

        from femsolver.elmer.equations.heat import ViewProxy

        self.assertEqual(ViewProxy, doc.Heat.ViewObject.Proxy.__class__)

        self.assertEqual(
            "Fem::EquationElmerMagnetodynamic2D",
            type_of_obj(ObjectsFem.makeEquationMagnetodynamic2D(doc)),
        )

        self.assertEqual(
            "Fem::EquationElmerMagnetodynamic",
            type_of_obj(ObjectsFem.makeEquationMagnetodynamic(doc)),
        )


"""
# code was generated by the following code from a document with all objects
# run test_object.test_femobjects_make how to create such document
# in tmp in FEM_unittests will be the file with all objects
# the doc.Name of some objects needs to be edited after run
# because obj. name != proxy type
# elmer equation objects need to be edited after
# material solid and material fluid

#view providers
from femtools.femutils import type_of_obj
for o in App.ActiveDocument.Objects:
    if hasattr(o, "Proxy"):
        vp_module_with_class = str(o.ViewObject.Proxy.__class__).lstrip("<class '").rstrip("'>")
        vp_class_of_o = vp_module_with_class.split(".")[-1]
        o_name_from_type = type_of_obj(o).lstrip('Fem::')
        vp_module_to_load = vp_module_with_class.rstrip(vp_class_of_o).rstrip(".")
        print("        from {} import {}".format(vp_module_to_load, vp_class_of_o))
        print("        self.assertEqual(")
        print("            {},".format(vp_class_of_o))
        print("            doc.{}.ViewObject.Proxy.__class__".format(o_name_from_type))
        print("        )")
        print("")

for o in App.ActiveDocument.Objects:
    if hasattr(o, "Proxy"):
        o.Proxy.__class__

"""



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

from dataclasses import dataclass
from enum import Enum, auto
import os
from typing import Optional

from addonmanager_freecad_interface import translate

from PySideWrapper import QtCore, QtWidgets

from .addonmanager_widget_addon_buttons import WidgetAddonButtons
from .addonmanager_widget_readme_browser import WidgetReadmeBrowser
from .addonmanager_colors import (
    warning_color_string,
    attention_color_string,
    bright_color_string,
)


class MessageType(Enum):
    Message = auto()
    Warning = auto()
    Error = auto()


@dataclass
class UpdateInformation:
    unchecked: bool = True
    check_in_progress: bool = False
    update_available: bool = False
    detached_head: bool = False
    version: str = ""
    tag: str = ""
    branch: Optional[str] = None


@dataclass
class WarningFlags:
    obsolete: bool = False
    python2: bool = False
    required_freecad_version: Optional[str] = None
    non_osi_approved: bool = False
    non_fsf_libre: bool = False


class PackageDetailsView(QtWidgets.QWidget):
    """The view class for the package details"""

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.button_bar = None
        self.readme_browser = None
        self.message_label = None
        self.location_label = None
        self.url_label = None
        self.installed = False
        self.disabled = False
        self.update_info = UpdateInformation()
        self.warning_flags = WarningFlags()
        self.installed_version = None
        self.installed_branch = None
        self.installed_timestamp = None
        self.can_disable = True
        self._setup_ui()

    def _setup_ui(self):
        self.vertical_layout = QtWidgets.QVBoxLayout(self)
        self.button_bar = WidgetAddonButtons(self)
        self.readme_browser = WidgetReadmeBrowser(self)
        self.message_label = QtWidgets.QLabel(self)
        self.location_label = QtWidgets.QLabel(self)
        self.url_label = QtWidgets.QLabel(self)
        self.url_label.setOpenExternalLinks(True)
        self.location_label.setOpenExternalLinks(True)
        self.vertical_layout.addWidget(self.button_bar)
        self.vertical_layout.addWidget(self.message_label)
        self.vertical_layout.addWidget(self.location_label)
        self.vertical_layout.addWidget(self.url_label)
        self.vertical_layout.addWidget(self.readme_browser)
        self.button_bar.hide()  # Start with no bar

    def set_location(self, location: Optional[str]):
        if location is not None:
            location_path = os.path.normpath(location)
            text = (
                f"{translate('AddonsInstaller', 'Installation location')}: "
                f'<a href="file://{location_path}">{location_path}</a>'
            )
            self.location_label.setText(text)
            self.location_label.show()
        else:
            self.location_label.hide()

    def set_url(self, url: Optional[str]):
        if url is not None:
            text = (
                translate("AddonsInstaller", "Repository URL")
                + ': <a href="'
                + url
                + '">'
                + url
                + "</a>"
            )
            self.url_label.setText(text)
            self.url_label.show()
        else:
            self.url_label.hide()

    def set_installed(
        self,
        installed: bool,
        on_date: float = None,
        version: Optional[str] = None,
        branch: Optional[str] = None,
    ):
        self.installed = installed
        self.installed_timestamp = on_date
        self.installed_version = version
        self.installed_branch = branch
        if not self.installed:
            self.set_location(None)
        self._sync_ui_state()

    def set_update_available(self, info: UpdateInformation):
        self.update_info = info
        self._sync_ui_state()

    def set_disabled(self, disabled: bool):
        self.disabled = disabled
        self._sync_ui_state()

    def allow_disabling(self, allow: bool):
        self.can_disable = allow
        self._sync_ui_state()

    def allow_running(self, show: bool):
        self.button_bar.run_macro.setVisible(show)

    def set_warning_flags(self, flags: WarningFlags):
        self.warning_flags = flags
        self._sync_ui_state()

    def set_new_disabled_status(self, disabled: bool):
        """If the user just changed the enabled/disabled state of the addon, display a message
        indicating that will not take place until restart. Do not call except in a case of a
        state change during this run."""

        if disabled:
            message = translate(
                "AddonsInstaller",
                "This addon will be disabled when restarting FreeCAD",
            )
        else:
            message = translate(
                "AddonsInstaller",
                "This addon will be enabled when restarting FreeCAD",
            )
        self.message_label.setText(f"<h3>{message}</h3>")
        self.message_label.setStyleSheet("color:" + attention_color_string())

    def set_new_branch(self, branch: str):
        """If the user just changed branches, update the message to show that a restart is
        needed."""
        message_string = "<h3>"
        message_string += translate(
            "AddonsInstaller", "Changed to branch '{}' -- restart FreeCAD to use the addon"
        ).format(branch)
        message_string += "</h3>"
        self.message_label.setText(message_string)
        self.message_label.setStyleSheet("color:" + attention_color_string())

    def set_updated(self):
        """If the user has just updated the addon but not yet restarted, show an indication that
        we are awaiting a restart."""
        message = translate(
            "AddonsInstaller",
            "This addon has been updated. Restart FreeCAD to see changes.",
        )
        self.message_label.setText(f"<h3>{message}</h3>")
        self.message_label.setStyleSheet("color:" + attention_color_string())

    def _sync_ui_state(self):
        self._create_status_label_text()

    def _create_status_label_text(self):
        if self.installed:
            installation_details = self._get_installation_details_string()
            update_details = self._get_update_status_string()
            message_text = f"{installation_details} {update_details}"
            if self.disabled:
                message_text += " [" + translate("AddonsInstaller", "Disabled") + "]"
            self.message_label.setText(f"<h3>{message_text}</h3>")
            if self.disabled:
                self.message_label.setStyleSheet("color:" + warning_color_string())
            elif self.update_info.update_available:
                self.message_label.setStyleSheet("color:" + attention_color_string())
            else:
                self.message_label.setStyleSheet("color:" + bright_color_string())
            self.message_label.show()
        elif self._there_are_warnings_to_show():
            warnings = self._get_warning_string()
            self.message_label.setText(f"<h3>{warnings}</h3>")
            self.message_label.setStyleSheet("color:" + warning_color_string())
            self.message_label.show()
        else:
            self.message_label.hide()

    def _get_installation_details_string(self) -> str:
        version = self.installed_version
        date = ""
        installed_version_string = ""
        if self.installed_timestamp:
            date = QtCore.QLocale().toString(
                QtCore.QDateTime.fromSecsSinceEpoch(int(round(self.installed_timestamp, 0))),
                QtCore.QLocale.ShortFormat,
            )
        if version and date:
            installed_version_string += (
                translate("AddonsInstaller", "Version {version} installed on {date}").format(
                    version=version, date=date
                )
                + ". "
            )
        elif version:
            installed_version_string += (
                translate("AddonsInstaller", "Version {version} installed") + "."
            ).format(version=version)
        elif date:
            installed_version_string += (
                translate("AddonsInstaller", "Installed on {date}") + "."
            ).format(date=date)
        else:
            installed_version_string += translate("AddonsInstaller", "Installed") + "."
        return installed_version_string

    def _get_update_status_string(self) -> str:
        if self.update_info.check_in_progress:
            return translate("AddonsInstaller", "Update check in progress") + "."
        elif self.update_info.unchecked:
            return ""
        if self.update_info.detached_head:
            return (
                translate(
                    "AddonsInstaller", "Git tag '{}' checked out, no updates possible"
                ).format(self.update_info.tag)
                + "."
            )
        if self.update_info.update_available:
            if self.installed_branch and self.update_info.branch:
                if self.installed_branch != self.update_info.branch:
                    return (
                        translate(
                            "AddonsInstaller",
                            "Currently on branch {}, name changed to {}",
                        ).format(self.installed_branch, self.update_info.branch)
                        + "."
                    )
                if self.update_info.version:
                    return (
                        translate(
                            "AddonsInstaller",
                            "Currently on branch {}, update available to version {}",
                        ).format(self.installed_branch, str(self.update_info.version).strip())
                        + "."
                    )
                return translate("AddonsInstaller", "Update available") + "."
            if self.update_info.version:
                return (
                    translate("AddonsInstaller", "Update available to version {}").format(
                        str(self.update_info.version).strip()
                    )
                    + "."
                )
            return translate("AddonsInstaller", "Update available") + "."
        return translate("AddonsInstaller", "This is the latest version available") + "."

    def _there_are_warnings_to_show(self) -> bool:
        if self.disabled:
            return True
        if self.warning_flags.required_freecad_version:
            return True
        return False  # TODO: Someday support optional warnings on license types

    def _get_warning_string(self) -> str:
        if self.installed and self.disabled:
            return translate(
                "AddonsInstaller",
                "WARNING: This addon is currently installed, but disabled. Use the 'enable' "
                "button to re-enable.",
            )
        if self.warning_flags.required_freecad_version:
            return translate("AddonsInstaller", "WARNING: This addon requires FreeCAD {}").format(
                self.warning_flags.required_freecad_version
            )
        return ""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore
import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.Util as PathGuiUtil
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Surface as PathSurface
import PathGui


__title__ = "CAM Surface Operation UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Surface operation page controller and command implementation."

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class TaskPanelOpPage(PathOpGui.TaskPanelPage):
    """Page controller class for the Surface operation."""

    def initPage(self, obj):
        self.setTitle("3D Surface - " + obj.Label)
        # self.updateVisibility()
        # retrieve property enumerations
        # self.propEnums = PathSurface.ObjectSurface.opPropertyEnumerations(False)
        self.propEnums = PathSurface.ObjectSurface.propertyEnumerations(False)

    def getForm(self):
        """getForm() ... returns UI"""
        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpSurfaceEdit.ui")
        comboToPropertyMap = [
            ("boundBoxSelect", "BoundBox"),
            ("scanType", "ScanType"),
            ("cutPattern", "CutPattern"),
            ("profileEdges", "ProfileEdges"),
            ("layerMode", "LayerMode"),
            ("dropCutterDirSelect", "DropCutterDir"),
        ]
        enumTups = PathSurface.ObjectSurface.propertyEnumerations(dataType="raw")
        PathGuiUtil.populateCombobox(form, enumTups, comboToPropertyMap)

        return form

    def getFields(self, obj):
        """getFields(obj) ... transfers values from UI to obj's properties"""
        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)

        if obj.BoundBox != str(self.form.boundBoxSelect.currentData()):
            obj.BoundBox = str(self.form.boundBoxSelect.currentData())

        if obj.ScanType != str(self.form.scanType.currentData()):
            obj.ScanType = str(self.form.scanType.currentData())

        if obj.LayerMode != str(self.form.layerMode.currentData()):
            obj.LayerMode = str(self.form.layerMode.currentData())

        """
        The following method of getting values from the UI form
            allows for translations of combobox options in the UI.
        The requirement is that the enumeration lists must
            be in the same order in both the opPropertyEnumerations() method
            and the UI panel QComboBox list.
        Another step to ensure synchronization of the two lists is to
            populate the list dynamically in this Gui module in `initPage()`
            using the property enumerations list when loading the UI panel.
            This type of dynamic combobox population is done for the
            Tool Controller selection.
        """
        # val = self.propEnums["CutPattern"][self.form.cutPattern.currentIndex()]
        # if obj.CutPattern != val:
        #     obj.CutPattern = val

        # val = self.propEnums["ProfileEdges"][self.form.profileEdges.currentIndex()]
        # if obj.ProfileEdges != val:
        #     obj.ProfileEdges = val

        obj.CutPattern = self.form.cutPattern.currentData()
        obj.ProfileEdges = self.form.profileEdges.currentData()

        if obj.AvoidLastX_Faces != self.form.avoidLastX_Faces.value():
            obj.AvoidLastX_Faces = self.form.avoidLastX_Faces.value()

        obj.DropCutterExtraOffset.x = FreeCAD.Units.Quantity(
            self.form.boundBoxExtraOffsetX.text()
        ).Value
        obj.DropCutterExtraOffset.y = FreeCAD.Units.Quantity(
            self.form.boundBoxExtraOffsetY.text()
        ).Value

        if obj.DropCutterDir != str(self.form.dropCutterDirSelect.currentData()):
            obj.DropCutterDir = str(self.form.dropCutterDirSelect.currentData())

        PathGuiUtil.updateInputField(obj, "DepthOffset", self.form.depthOffset)

        if obj.StepOver != self.form.stepOver.value():
            obj.StepOver = self.form.stepOver.value()

        PathGuiUtil.updateInputField(obj, "SampleInterval", self.form.sampleInterval)

        if obj.UseStartPoint != self.form.useStartPoint.isChecked():
            obj.UseStartPoint = self.form.useStartPoint.isChecked()

        if obj.BoundaryEnforcement != self.form.boundaryEnforcement.isChecked():
            obj.BoundaryEnforcement = self.form.boundaryEnforcement.isChecked()

        if obj.OptimizeLinearPaths != self.form.optimizeEnabled.isChecked():
            obj.OptimizeLinearPaths = self.form.optimizeEnabled.isChecked()

        if obj.OptimizeStepOverTransitions != self.form.optimizeStepOverTransitions.isChecked():
            obj.OptimizeStepOverTransitions = self.form.optimizeStepOverTransitions.isChecked()

    def setFields(self, obj):
        """setFields(obj) ... transfers obj's property values to UI"""
        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)
        self.selectInComboBox(obj.BoundBox, self.form.boundBoxSelect)
        self.selectInComboBox(obj.ScanType, self.form.scanType)
        self.selectInComboBox(obj.LayerMode, self.form.layerMode)

        """
        The following method of setting values in the UI form
            allows for translations of combobox options in the UI.
        The requirement is that the enumeration lists must
            be in the same order in both the opPropertyEnumerations() method
            and the UI panel QComboBox list.
        The original method is commented out below.
        """
        # idx = self.propEnums["CutPattern"].index(obj.CutPattern)
        # self.form.cutPattern.setCurrentIndex(idx)
        # idx = self.propEnums["ProfileEdges"].index(obj.ProfileEdges)
        # self.form.profileEdges.setCurrentIndex(idx)
        self.selectInComboBox(obj.CutPattern, self.form.cutPattern)
        self.selectInComboBox(obj.ProfileEdges, self.form.profileEdges)

        self.form.avoidLastX_Faces.setValue(obj.AvoidLastX_Faces)
        self.form.boundBoxExtraOffsetX.setText(
            FreeCAD.Units.Quantity(obj.DropCutterExtraOffset.x, FreeCAD.Units.Length).UserString
        )
        self.form.boundBoxExtraOffsetY.setText(
            FreeCAD.Units.Quantity(obj.DropCutterExtraOffset.y, FreeCAD.Units.Length).UserString
        )
        self.selectInComboBox(obj.DropCutterDir, self.form.dropCutterDirSelect)
        self.form.depthOffset.setText(
            FreeCAD.Units.Quantity(obj.DepthOffset.Value, FreeCAD.Units.Length).UserString
        )
        self.form.stepOver.setValue(obj.StepOver)
        self.form.sampleInterval.setText(
            FreeCAD.Units.Quantity(obj.SampleInterval.Value, FreeCAD.Units.Length).UserString
        )

        if obj.UseStartPoint:
            self.form.useStartPoint.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.useStartPoint.setCheckState(QtCore.Qt.Unchecked)

        if obj.BoundaryEnforcement:
            self.form.boundaryEnforcement.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.boundaryEnforcement.setCheckState(QtCore.Qt.Unchecked)

        if obj.OptimizeLinearPaths:
            self.form.optimizeEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.optimizeEnabled.setCheckState(QtCore.Qt.Unchecked)

        if obj.OptimizeStepOverTransitions:
            self.form.optimizeStepOverTransitions.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.optimizeStepOverTransitions.setCheckState(QtCore.Qt.Unchecked)

        self.updateVisibility()

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.coolantController.currentIndexChanged)
        signals.append(self.form.boundBoxSelect.currentIndexChanged)
        signals.append(self.form.scanType.currentIndexChanged)
        signals.append(self.form.layerMode.currentIndexChanged)
        signals.append(self.form.cutPattern.currentIndexChanged)
        signals.append(self.form.profileEdges.currentIndexChanged)
        signals.append(self.form.avoidLastX_Faces.editingFinished)
        signals.append(self.form.boundBoxExtraOffsetX.editingFinished)
        signals.append(self.form.boundBoxExtraOffsetY.editingFinished)
        signals.append(self.form.dropCutterDirSelect.currentIndexChanged)
        signals.append(self.form.depthOffset.editingFinished)
        signals.append(self.form.stepOver.editingFinished)
        signals.append(self.form.sampleInterval.editingFinished)
        if hasattr(self.form.useStartPoint, "checkStateChanged"):  # Qt version >= 6.7.0
            signals.append(self.form.useStartPoint.checkStateChanged)
            signals.append(self.form.boundaryEnforcement.checkStateChanged)
            signals.append(self.form.optimizeEnabled.checkStateChanged)
            signals.append(self.form.optimizeStepOverTransitions.checkStateChanged)
        else:  # Qt version < 6.7.0
            signals.append(self.form.useStartPoint.stateChanged)
            signals.append(self.form.boundaryEnforcement.stateChanged)
            signals.append(self.form.optimizeEnabled.stateChanged)
            signals.append(self.form.optimizeStepOverTransitions.stateChanged)

        return signals

    def updateVisibility(self, sentObj=None):
        """updateVisibility(sentObj=None)... Updates visibility of Tasks panel objects."""
        if self.form.scanType.currentText() == "Planar":
            self.form.cutPattern.show()
            self.form.cutPattern_label.show()
            self.form.optimizeStepOverTransitions.show()
            if hasattr(self.form, "profileEdges"):
                self.form.profileEdges.show()
                self.form.profileEdges_label.show()
                self.form.avoidLastX_Faces.show()
                self.form.avoidLastX_Faces_label.show()

            self.form.boundBoxExtraOffsetX.hide()
            self.form.boundBoxExtraOffsetY.hide()
            self.form.boundBoxExtraOffset_label.hide()
            self.form.dropCutterDirSelect.hide()
            self.form.dropCutterDirSelect_label.hide()
        elif self.form.scanType.currentText() == "Rotational":
            self.form.cutPattern.hide()
            self.form.cutPattern_label.hide()
            self.form.optimizeStepOverTransitions.hide()
            if hasattr(self.form, "profileEdges"):
                self.form.profileEdges.hide()
                self.form.profileEdges_label.hide()
                self.form.avoidLastX_Faces.hide()
                self.form.avoidLastX_Faces_label.hide()

            self.form.boundBoxExtraOffsetX.show()
            self.form.boundBoxExtraOffsetY.show()
            self.form.boundBoxExtraOffset_label.show()
            self.form.dropCutterDirSelect.show()
            self.form.dropCutterDirSelect_label.show()

    def registerSignalHandlers(self, obj):
        self.form.scanType.currentIndexChanged.connect(self.updateVisibility)


Command = PathOpGui.SetupOperation(
    "Surface",
    PathSurface.Create,
    TaskPanelOpPage,
    "CAM_3DSurface",
    QtCore.QT_TRANSLATE_NOOP("CAM_Surface", "3D Surface"),
    QtCore.QT_TRANSLATE_NOOP("CAM_Surface", "Create a 3D Surface Operation from a model"),
    PathSurface.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathSurfaceGui... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Defines the PackageList QWidget for displaying a list of Addons."""
import threading

import addonmanager_freecad_interface as fci
from PySideWrapper import QtCore, QtGui, QtWidgets

from Addon import Addon

from compact_view import Ui_CompactView
from expanded_view import Ui_ExpandedView

import addonmanager_utilities as utils
from Widgets.addonmanager_widget_view_control_bar import WidgetViewControlBar, SortOptions
from Widgets.addonmanager_widget_view_selector import AddonManagerDisplayStyle
from Widgets.addonmanager_widget_filter_selector import StatusFilter, Filter
from Widgets.addonmanager_widget_progress_bar import Progress, WidgetProgressBar
from addonmanager_licenses import get_license_manager

translate = fci.translate


# pylint: disable=too-few-public-methods


class PackageList(QtWidgets.QWidget):
    """A widget that shows a list of packages and various widgets to control the
    display of the list, including a progress bar that can display and interrupt the load
    process."""

    itemSelected = QtCore.Signal(Addon)
    stop_loading = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_PackageList()
        self.ui.setupUi(self)

        self.item_filter = PackageListFilter()
        self.ui.listPackages.setModel(self.item_filter)
        self.item_delegate = PackageListItemDelegate(self.ui.listPackages)
        self.ui.listPackages.setItemDelegate(self.item_delegate)

        self.ui.listPackages.clicked.connect(self.on_listPackages_clicked)
        self.ui.view_bar.filter_changed.connect(self.update_status_filter)
        self.ui.view_bar.search_changed.connect(self.item_filter.setFilterRegularExpression)
        self.ui.view_bar.sort_changed.connect(self.item_filter.setSortRole)
        self.ui.view_bar.sort_changed.connect(self.item_delegate.set_sort)
        self.ui.view_bar.sort_order_changed.connect(lambda order: self.item_filter.sort(0, order))
        self.ui.progress_bar.stop_clicked.connect(self.stop_loading)

        # Set up the view the same as the last time:
        package_type = fci.Preferences().get("PackageTypeSelection")
        status = fci.Preferences().get("StatusSelection")
        search_string = fci.Preferences().get("SearchString")
        self.ui.view_bar.filter_selector.set_contents_filter(package_type)
        self.ui.view_bar.filter_selector.set_status_filter(status)
        if search_string:
            self.ui.view_bar.search.filter_line_edit.setText(search_string)
        self.item_filter.setPackageFilter(package_type)
        self.item_filter.setStatusFilter(status)

        # Pre-init of other members:
        self.item_model = None

    def setModel(self, model):
        """This is a model-view-controller widget: set its model."""
        self.item_model = model
        self.item_filter.setSourceModel(self.item_model)
        self.item_filter.setSortRole(SortOptions.Alphabetical)
        self.item_filter.sort(0, QtCore.Qt.AscendingOrder)

        style = fci.Preferences().get("ViewStyle")
        self.set_view_style(style)
        self.ui.view_bar.view_selector.set_current_view(style)

        self.item_filter.setHideNonOSIApproved(fci.Preferences().get("HideNonOSIApproved"))
        self.item_filter.setHideNonFSFLibre(fci.Preferences().get("HideNonFSFFreeLibre"))
        self.item_filter.setHideUnlicensed(fci.Preferences().get("HideUnlicensed"))

    def select_addon(self, addon_name: str):
        for index, addon in enumerate(self.item_model.repos):
            if addon.name == addon_name:
                row_index = self.item_model.createIndex(index, 0)
                if self.item_filter.filterAcceptsRow(index):
                    self.ui.listPackages.setCurrentIndex(row_index)
                else:
                    fci.Console.PrintLog(
                        f"Addon {addon_name} is not visible given current "
                        "filter: not selecting it."
                    )
                return
        fci.Console.PrintLog(f"Could not find addon '{addon_name}' to select it")

    def on_listPackages_clicked(self, index: QtCore.QModelIndex):
        """Determine what addon was selected and emit the itemSelected signal with it as
        an argument."""
        source_selection = self.item_filter.mapToSource(index)
        selected_repo = self.item_model.repos[source_selection.row()]
        self.itemSelected.emit(selected_repo)

    def update_status_filter(self, new_filter: Filter) -> None:
        """hide/show rows corresponding to the specified filter"""

        self.item_filter.setStatusFilter(new_filter.status_filter)
        self.item_filter.setPackageFilter(new_filter.content_filter)
        fci.Preferences().set("StatusSelection", new_filter.status_filter)
        fci.Preferences().set("PackageTypeSelection", new_filter.content_filter)
        self.item_filter.invalidateFilter()

    def set_view_style(self, style: AddonManagerDisplayStyle) -> None:
        """Set the style (compact or expanded) of the list"""
        if self.item_model:
            self.item_model.layoutAboutToBeChanged.emit()
        self.item_delegate.set_view(style)
        if style == AddonManagerDisplayStyle.COMPACT or style == AddonManagerDisplayStyle.COMPOSITE:
            self.ui.listPackages.setSpacing(2)
            self.ui.listPackages.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerItem)
            self.ui.listPackages.verticalScrollBar().setSingleStep(-1)
        else:
            self.ui.listPackages.setSpacing(5)
            self.ui.listPackages.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
            self.ui.listPackages.verticalScrollBar().setSingleStep(24)
        if self.item_model:
            self.item_model.layoutChanged.emit()

        fci.Preferences().set("ViewStyle", style)

    def set_loading(self, is_loading: bool) -> None:
        """Set the loading status of this package list: when a package list is loading, it shows
        a progress bar. When it is no longer loading, the bar is hidden and the search bar gets
        the focus."""
        if is_loading:
            self.ui.progress_bar.show()
        else:
            self.ui.progress_bar.hide()
            self.ui.view_bar.search.setFocus()

    def update_loading_progress(self, progress: Progress) -> None:
        self.ui.progress_bar.set_progress(progress)


class PackageListItemModel(QtCore.QAbstractListModel):
    """The model for use with the PackageList class."""

    repos = []
    write_lock = threading.Lock()

    DataAccessRole = QtCore.Qt.UserRole

    def rowCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        """The number of rows"""
        if parent.isValid():
            return 0
        return len(self.repos)

    def columnCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        """Only one column, always returns 1."""
        if parent.isValid():
            return 0
        return 1

    def data(self, index: QtCore.QModelIndex, role: int = QtCore.Qt.DisplayRole):
        """Get the data for a given index and role."""
        if not index.isValid():
            return None
        row = index.row()
        if role == QtCore.Qt.ToolTipRole:
            tooltip = ""
            if self.repos[row].repo_type == Addon.Kind.PACKAGE:
                tooltip = translate("AddonsInstaller", "Click for details about package {}").format(
                    self.repos[row].display_name
                )
            elif self.repos[row].repo_type == Addon.Kind.WORKBENCH:
                tooltip = translate(
                    "AddonsInstaller", "Click for details about workbench {}"
                ).format(self.repos[row].display_name)
            elif self.repos[row].repo_type == Addon.Kind.MACRO:
                tooltip = translate("AddonsInstaller", "Click for details about macro {}").format(
                    self.repos[row].display_name
                )
            return tooltip
        if role == PackageListItemModel.DataAccessRole:
            return self.repos[row]

        # Sorting
        if role == SortOptions.Alphabetical:
            return self.repos[row].display_name
        if role == SortOptions.LastUpdated:
            update_date = self.repos[row].update_date
            if update_date and hasattr(update_date, "timestamp"):
                return update_date.timestamp()
            return 0
        if role == SortOptions.DateAdded:
            if self.repos[row].stats and self.repos[row].stats.date_created:
                return self.repos[row].stats.date_created.timestamp()
            return 0
        if role == SortOptions.Stars:
            if self.repos[row].stats and self.repos[row].stats.stars:
                return self.repos[row].stats.stars
            return 0
        if role == SortOptions.Score:
            return self.repos[row].score

        # Anything else:
        return None

    def headerData(self, _unused1, _unused2, _role=QtCore.Qt.DisplayRole):
        """No header in this implementation: always returns None."""
        return None

    def append_item(self, repo: Addon) -> None:
        """Adds this addon to the end of the model. Thread safe."""
        if repo in self.repos:
            # Cowardly refuse to insert the same repo a second time
            return
        with self.write_lock:
            self.beginInsertRows(QtCore.QModelIndex(), self.rowCount(), self.rowCount())
            self.repos.append(repo)
            self.endInsertRows()

    def remove_item(self, repo: Addon) -> None:
        if repo not in self.repos:
            # Nothing to remove, don't care
            return
        with self.write_lock:
            self.beginRemoveRows(
                QtCore.QModelIndex(), self.repos.index(repo), self.repos.index(repo)
            )
            self.repos.remove(repo)
            self.endRemoveRows()

    def clear(self) -> None:
        """Clear the model, removing all rows. Thread safe."""
        if self.rowCount() > 0:
            with self.write_lock:
                self.beginRemoveRows(QtCore.QModelIndex(), 0, self.rowCount() - 1)
                self.repos = []
                self.endRemoveRows()

    def reload_item(self, repo: Addon) -> None:
        """Sets the addon data for the given addon (based on its name)"""
        for index, item in enumerate(self.repos):
            if item.name == repo.name:
                with self.write_lock:
                    self.repos[index] = repo
                return


class CompactView(QtWidgets.QWidget):
    """A single-line view of the package information"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_CompactView()
        self.ui.setupUi(self)


class ExpandedView(QtWidgets.QWidget):
    """A multi-line view of the package information"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_ExpandedView()
        self.ui.setupUi(self)


class PackageListItemDelegate(QtWidgets.QStyledItemDelegate):
    """Render the repo data as a formatted region"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.displayStyle = AddonManagerDisplayStyle.EXPANDED
        self.sort_order = SortOptions.Alphabetical
        self.expanded = ExpandedView()
        self.compact = CompactView()
        self.widget = self.expanded

    def set_view(self, style: AddonManagerDisplayStyle) -> None:
        """Set the view of to style"""
        if not self.displayStyle == style:
            self.displayStyle = style

    def set_sort(self, sort: SortOptions) -> None:
        """When sorting by various things, we display the thing that's being sorted on."""
        if not self.sort_order == sort:
            self.sort_order = sort

    def sizeHint(self, _option, index):
        """Attempt to figure out the correct height for the widget based on its
        current contents."""
        self.update_content(index)
        return self.widget.sizeHint()

    def update_content(self, index):
        """Creates the display of the content for a given index."""
        repo = index.data(PackageListItemModel.DataAccessRole)
        if self.displayStyle == AddonManagerDisplayStyle.EXPANDED:
            self.widget = self.expanded
            self._setup_expanded_view(repo)
        elif self.displayStyle == AddonManagerDisplayStyle.COMPACT:
            self.widget = self.compact
            self._setup_compact_view(repo)
        elif self.displayStyle == AddonManagerDisplayStyle.COMPOSITE:
            self.widget = self.compact  # For now reuse the compact list
            self._setup_composite_view(repo)
        self.widget.adjustSize()

    def _setup_expanded_view(self, addon: Addon) -> None:
        self.widget.ui.labelPackageName.setText(f"<h1>{addon.display_name}</h1>")
        self.widget.ui.labelIcon.setPixmap(addon.icon.pixmap(QtCore.QSize(48, 48)))
        self.widget.ui.labelStatus.setText(self.get_expanded_update_string(addon))
        self.widget.ui.labelIcon.setText("")
        self.widget.ui.labelTags.setText("")
        if addon.metadata:
            self.widget.ui.labelDescription.setText(addon.metadata.description)
            self.widget.ui.labelVersion.setText(f"<i>v{addon.metadata.version}</i>")
            self._set_package_maintainer_label(addon)
        elif addon.macro:
            self.widget.ui.labelDescription.setText(addon.macro.comment)
            self._set_macro_version_label(addon)
            self._set_macro_maintainer_label(addon)
        else:
            self.widget.ui.labelDescription.setText("")
            self.widget.ui.labelMaintainer.setText("")
            self.widget.ui.labelVersion.setText("")
        if addon.tags:
            self.widget.ui.labelTags.setText(
                translate("AddonsInstaller", "Tags") + ": " + ", ".join(addon.tags)
            )
        if self.sort_order == SortOptions.Alphabetical:
            self.widget.ui.labelSort.setText("")
        else:
            self.widget.ui.labelSort.setText(self._get_sort_label_text(addon))

    def _setup_compact_view(self, addon: Addon) -> None:
        self.widget.ui.labelPackageName.setText(f"<b>{addon.display_name}</b>")
        self.widget.ui.labelIcon.setPixmap(addon.icon.pixmap(QtCore.QSize(16, 16)))
        self.widget.ui.labelStatus.setText(self.get_compact_update_string(addon))
        self.widget.ui.labelIcon.setText("")
        if addon.metadata:
            self.widget.ui.labelVersion.setText(f"<i>v{addon.metadata.version}</i>")
        elif addon.macro:
            self._set_macro_version_label(addon)
        else:
            self.widget.ui.labelVersion.setText("")
        if self.sort_order == SortOptions.Alphabetical:
            description = self._get_compact_description(addon)
            self.widget.ui.labelDescription.setText(description)
        else:
            self.widget.ui.labelDescription.setText(self._get_sort_label_text(addon))

    def _setup_composite_view(self, addon: Addon) -> None:
        self.widget.ui.labelPackageName.setText(f"<b>{addon.display_name}</b>")
        self.widget.ui.labelIcon.setPixmap(addon.icon.pixmap(QtCore.QSize(16, 16)))
        self.widget.ui.labelStatus.setText(self.get_compact_update_string(addon))
        self.widget.ui.labelIcon.setText("")
        if addon.metadata:
            self.widget.ui.labelVersion.setText(f"<i>v{addon.metadata.version}</i>")
        elif addon.macro:
            self._set_macro_version_label(addon)
        else:
            self.widget.ui.labelVersion.setText("")
        if self.sort_order != SortOptions.Alphabetical:
            self.widget.ui.labelDescription.setText(self._get_sort_label_text(addon))
        else:
            self.widget.ui.labelDescription.setText("")

    def _set_package_maintainer_label(self, addon: Addon):

        maintainers = addon.metadata.maintainer

        text = ""

        count = len(maintainers)

        if count > 0:

            text = translate("AddonsInstaller", "Maintainer(s)", "", count)
            text = f"{ text }: "

            for maintainer in maintainers:

                if count > 1:
                    text += "\n"

                text += maintainer.name

                if maintainer.email:
                    text += f" <{ maintainer.email }>"

        self.widget.ui.labelMaintainer.setText(text)

    def _set_macro_maintainer_label(self, repo: Addon):
        if repo.macro.author:
            caption = translate("AddonsInstaller", "Author")
            self.widget.ui.labelMaintainer.setText(caption + ": " + repo.macro.author)
        else:
            self.widget.ui.labelMaintainer.setText("")

    def _set_macro_version_label(self, addon: Addon):
        version_string = ""
        if addon.macro.version:
            version_string = addon.macro.version + " "
        if addon.macro.on_wiki:
            version_string += "(wiki)"
        elif addon.macro.on_git:
            version_string += "(git)"
        else:
            version_string += "(unknown source)"
        self.widget.ui.labelVersion.setText("<i>" + version_string + "</i>")

    def _get_sort_label_text(self, addon: Addon) -> str:
        if self.sort_order == SortOptions.Alphabetical:
            return ""
        elif self.sort_order == SortOptions.Stars:
            if addon.stats and addon.stats.stars and addon.stats.stars > 0:
                return translate("AddonsInstaller", "{} ★ on GitHub").format(addon.stats.stars)
            return translate("AddonsInstaller", "No ★, or not on GitHub")
        elif self.sort_order == SortOptions.DateAdded:
            if addon.stats and addon.stats.date_created:
                epoch_seconds = addon.stats.date_created.timestamp()
                qdt = QtCore.QDateTime.fromSecsSinceEpoch(int(epoch_seconds)).date()
                time_string = QtCore.QLocale().toString(qdt, QtCore.QLocale.ShortFormat)
                return translate("AddonsInstaller", "Created ") + time_string
            return ""
        elif self.sort_order == SortOptions.LastUpdated:
            update_date = addon.update_date
            if update_date:
                epoch_seconds = update_date.timestamp()
                qdt = QtCore.QDateTime.fromSecsSinceEpoch(int(epoch_seconds)).date()
                time_string = QtCore.QLocale().toString(qdt, QtCore.QLocale.ShortFormat)
                return translate("AddonsInstaller", "Updated ") + time_string
            return ""
        elif self.sort_order == SortOptions.Score:
            return translate("AddonsInstaller", "Score: ") + str(addon.score)
        return ""

    def _get_compact_description(self, addon: Addon) -> str:
        description = ""
        if addon.metadata:
            description = addon.metadata.description
        elif addon.macro and addon.macro.comment:
            description = addon.macro.comment
        trimmed_text, _, _ = description.partition(".")
        return trimmed_text.replace("\n", " ")

    @staticmethod
    def get_compact_update_string(repo: Addon) -> str:
        """Get a single-line string listing details about the installed version and
        date"""

        result = ""
        if repo.status() == Addon.Status.UNCHECKED:
            result = translate("AddonsInstaller", "Installed")
        elif repo.status() == Addon.Status.NO_UPDATE_AVAILABLE:
            result = translate("AddonsInstaller", "Up-to-date")
        elif repo.status() == Addon.Status.UPDATE_AVAILABLE:
            result = translate("AddonsInstaller", "Update available")
        elif repo.status() == Addon.Status.PENDING_RESTART:
            result = translate("AddonsInstaller", "Pending restart")

        if repo.is_disabled():
            style = "style='color:" + utils.warning_color_string() + "; font-weight:bold;'"
            result += f"<span {style}> [" + translate("AddonsInstaller", "DISABLED") + "]</span>"

        return result

    @staticmethod
    def get_expanded_update_string(repo: Addon) -> str:
        """Get a multi-line string listing details about the installed version and
        date"""

        result = ""

        installed_version_string = ""
        if repo.status() != Addon.Status.NOT_INSTALLED:
            if repo.installed_version or repo.installed_metadata:
                installed_version_string = (
                    "<br/>" + translate("AddonsInstaller", "Installed version") + ": "
                )
                if repo.installed_metadata:
                    installed_version_string += str(repo.installed_metadata.version)
                elif repo.installed_version:
                    installed_version_string += str(repo.installed_version)
            else:
                installed_version_string = "<br/>" + translate("AddonsInstaller", "Unknown version")

        if repo.updated_timestamp:
            installed_date_string = QtCore.QLocale().toString(
                QtCore.QDateTime.fromSecsSinceEpoch(int(round(repo.updated_timestamp, 0))),
                QtCore.QLocale.ShortFormat,
            )
            installed_version_string += " (" + installed_date_string + ")"

        available_version_string = ""
        if repo.metadata:
            available_version_string = (
                "<br/>" + translate("AddonsInstaller", "Available version") + ": "
            )
            available_version_string += str(repo.metadata.version)

        remote_updated_date_string = ""
        if repo.remote_last_updated:
            py_dt = repo.remote_last_updated
            q_date = QtCore.QDate(py_dt.year, py_dt.month, py_dt.day)
            q_time = QtCore.QTime(py_dt.hour, py_dt.minute, py_dt.second, py_dt.microsecond // 1000)
            q_dt = QtCore.QDateTime(q_date, q_time)
            remote_updated_date_string = QtCore.QLocale().toString(
                q_dt,
                QtCore.QLocale.ShortFormat,
            )
            available_version_string += " (" + remote_updated_date_string + ")"

        if repo.status() == Addon.Status.UNCHECKED:
            result = translate("AddonsInstaller", "Installed")
            result += installed_version_string
        elif repo.status() == Addon.Status.NO_UPDATE_AVAILABLE:
            result = translate("AddonsInstaller", "Up-to-date")
            result += installed_version_string
        elif repo.status() == Addon.Status.UPDATE_AVAILABLE:
            result = translate("AddonsInstaller", "Update available")
            result += installed_version_string
            result += available_version_string
        elif repo.status() == Addon.Status.PENDING_RESTART:
            result = translate("AddonsInstaller", "Pending restart")

        if repo.is_disabled():
            style = "style='color:" + utils.warning_color_string() + "; font-weight:bold;'"
            result += (
                f"<br/><span {style}>[" + translate("AddonsInstaller", "DISABLED") + "]</span>"
            )

        return result

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        _: QtCore.QModelIndex,
    ):
        """Main paint function: renders this widget into a given rectangle,
        successively drawing all of its children."""
        painter.save()
        self.widget.resize(option.rect.size())
        painter.translate(option.rect.topLeft())
        self.widget.render(
            painter, QtCore.QPoint(), QtGui.QRegion(), QtWidgets.QWidget.DrawChildren
        )
        painter.restore()


class PackageListFilter(QtCore.QSortFilterProxyModel):
    """Handle filtering the item list on various criteria"""

    def __init__(self):
        super().__init__()
        self.package_type = 0  # Default to showing everything
        self.status = 0  # Default to showing any
        self.setSortCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.hide_non_OSI_approved = False
        self.hide_non_FSF_libre = False
        self.hide_unlicensed = False
        self.hide_newer_freecad_required = False

    def setPackageFilter(
        self, package_type: int
    ) -> None:  # 0=All, 1=Workbenches, 2=Macros, 3=Preference Packs, 4=Bundles, 5=Other
        """Set the package filter to package_type and refreshes."""
        self.package_type = package_type
        self.invalidateFilter()

    def setStatusFilter(
        self, status: int
    ) -> None:  # 0=Any, 1=Installed, 2=Not installed, 3=Update available
        """Sets the status filter to status and refreshes."""
        self.status = status
        self.invalidateFilter()

    def setHideNonOSIApproved(self, hide: bool) -> None:
        """Sets whether to hide Addons with non-OSI-approved licenses"""
        self.hide_non_OSI_approved = hide
        self.invalidateFilter()

    def setHideNonFSFLibre(self, hide: bool) -> None:
        """Sets whether to hide Addons with non-FSF-Libre licenses"""
        self.hide_non_FSF_libre = hide
        self.invalidateFilter()

    def setHideUnlicensed(self, hide: bool) -> None:
        """Sets whether to hide addons without a specified license"""
        self.hide_unlicensed = hide
        self.invalidateFilter()

    def filterAcceptsRow(self, row, _parent=QtCore.QModelIndex()):
        """Do the actual filtering (called automatically by Qt when drawing the list)"""

        index = self.sourceModel().createIndex(row, 0)
        data = self.sourceModel().data(index, PackageListItemModel.DataAccessRole)
        if self.package_type == 1:
            if not data.contains_workbench():
                return False
        elif self.package_type == 2:
            if not data.contains_macro():
                return False
        elif self.package_type == 3:
            if not data.contains_preference_pack():
                return False
        elif self.package_type == 4:
            if not data.contains_bundle():
                return False
        elif self.package_type == 5:
            if not data.contains_other():
                return False

        if self.status == StatusFilter.INSTALLED:
            if data.status() == Addon.Status.NOT_INSTALLED:
                return False
        elif self.status == StatusFilter.NOT_INSTALLED:
            if data.status() != Addon.Status.NOT_INSTALLED:
                return False
        elif self.status == StatusFilter.UPDATE_AVAILABLE:
            if data.status() != Addon.Status.UPDATE_AVAILABLE:
                return False

        license_manager = get_license_manager()
        if data.status() == Addon.Status.NOT_INSTALLED:

            if self.hide_unlicensed:
                if not data.license or data.license in ["UNLICENSED", "UNLICENCED"]:
                    fci.Console.PrintLog(f"Hiding {data.name} because it has no license set\n")
                    return False

            # If it is not an OSI-approved license, check to see if we are hiding those
            if self.hide_non_OSI_approved or self.hide_non_FSF_libre:
                if not data.license:
                    return False
                licenses_to_check = []
                if type(data.license) is str:
                    licenses_to_check.append(data.license)
                elif type(data.license) is list:
                    for license_id in data.license:
                        if type(license_id) is str:
                            licenses_to_check.append(license_id)
                        else:
                            licenses_to_check.append(license_id.name)
                else:
                    licenses_to_check.append(data.license.name)

                fsf_libre = False
                osi_approved = False
                for license_id in licenses_to_check:
                    if not osi_approved and license_manager.is_osi_approved(license_id):
                        osi_approved = True
                    if not fsf_libre and license_manager.is_fsf_libre(license_id):
                        fsf_libre = True
                if self.hide_non_OSI_approved and not osi_approved:
                    # fci.Console.PrintLog(
                    #    f"Hiding addon {data.name} because its license, {licenses_to_check}, "
                    #    f"is "
                    #    f"not OSI approved\n"
                    # )
                    return False
                if self.hide_non_FSF_libre and not fsf_libre:
                    # fci.Console.PrintLog(
                    #    f"Hiding addon {data.name} because its license, {licenses_to_check},  is "
                    #    f"not FSF Libre\n"
                    # )
                    return False

        name = data.display_name
        desc = data.description
        if hasattr(self, "filterRegularExpression"):  # Added in Qt 5.12
            re = self.filterRegularExpression()
            if re.isValid():
                re.setPatternOptions(QtCore.QRegularExpression.CaseInsensitiveOption)
                if re.match(name).hasMatch():
                    return True
                if re.match(desc).hasMatch():
                    return True
                if data.macro and data.macro.comment and re.match(data.macro.comment).hasMatch():
                    return True
                for tag in data.tags:
                    if re.match(tag).hasMatch():
                        return True
            return False
        # Only get here for Qt < 5.12
        re = self.filterRegExp()
        if re.isValid():
            re.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
            if re.indexIn(name) != -1:
                return True
            if re.indexIn(desc) != -1:
                return True
            if data.macro and data.macro.comment and re.indexIn(data.macro.comment) != -1:
                return True
            for tag in data.tags:
                if re.indexIn(tag) != -1:
                    return True
        return False


# pylint: disable=attribute-defined-outside-init, missing-function-docstring


class Ui_PackageList:
    """The contents of the PackageList widget"""

    def setupUi(self, form):
        if not form.objectName():
            form.setObjectName("PackageList")
        self.verticalLayout = QtWidgets.QVBoxLayout(form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")

        self.view_bar = WidgetViewControlBar(form)
        self.view_bar.setObjectName("ViewControlBar")
        self.horizontalLayout_6.addWidget(self.view_bar)

        self.verticalLayout.addLayout(self.horizontalLayout_6)

        self.listPackages = QtWidgets.QListView(form)
        self.listPackages.setObjectName("listPackages")
        self.listPackages.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.listPackages.setProperty("showDropIndicator", False)
        self.listPackages.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.listPackages.setResizeMode(QtWidgets.QListView.Adjust)
        self.listPackages.setUniformItemSizes(False)
        self.listPackages.setAlternatingRowColors(True)
        self.listPackages.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.listPackages.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.listPackages.verticalScrollBar().setSingleStep(24)

        self.verticalLayout.addWidget(self.listPackages)

        self.progress_bar = WidgetProgressBar()
        self.verticalLayout.addWidget(self.progress_bar)

        QtCore.QMetaObject.connectSlotsByName(form)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Used for CNC machine Stops for Path module. Create an Optional or Mandatory Stop."""

import FreeCAD
import FreeCADGui
import Path
from PySide import QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP

translate = FreeCAD.Qt.translate


class Stop:
    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyEnumeration",
            "Stop",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Add an optional or mandatory stop to the program"),
        )
        obj.Stop = ["Optional", "Mandatory"]
        obj.Proxy = self
        mode = 2
        obj.setEditorMode("Placement", mode)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, obj, prop):
        pass

    def execute(self, obj):
        if obj.Stop == "Optional":
            word = "M1"
        else:
            word = "M0"

        output = ""
        output = word + "\n"
        path = Path.Path(output)
        obj.Path = path


class _ViewProviderStop:
    def __init__(self, vobj):  # mandatory
        vobj.Proxy = self
        mode = 2
        vobj.setEditorMode("LineWidth", mode)
        vobj.setEditorMode("MarkerColor", mode)
        vobj.setEditorMode("NormalColor", mode)
        vobj.setEditorMode("DisplayMode", mode)
        vobj.setEditorMode("BoundingBox", mode)
        vobj.setEditorMode("Selectable", mode)
        vobj.setEditorMode("ShapeAppearance", mode)
        vobj.setEditorMode("Transparency", mode)
        vobj.setEditorMode("Visibility", mode)

    def dumps(self):  # mandatory
        return None

    def loads(self, state):  # mandatory
        return None

    def getIcon(self):  # optional
        return ":/icons/CAM_Stop.svg"

    def onChanged(self, vobj, prop):  # optional
        mode = 2
        vobj.setEditorMode("LineWidth", mode)
        vobj.setEditorMode("MarkerColor", mode)
        vobj.setEditorMode("NormalColor", mode)
        vobj.setEditorMode("DisplayMode", mode)
        vobj.setEditorMode("BoundingBox", mode)
        vobj.setEditorMode("Selectable", mode)
        vobj.setEditorMode("ShapeAppearance", mode)
        vobj.setEditorMode("Transparency", mode)
        vobj.setEditorMode("Visibility", mode)


class CommandPathStop:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Stop",
            "MenuText": QT_TRANSLATE_NOOP("CAM_Stop", "Stop"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_Stop", "Adds an optional or mandatory stop to the program"
            ),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        FreeCAD.ActiveDocument.openTransaction("Add Optional or Mandatory Stop to the program")
        FreeCADGui.addModule("Path.Op.Gui.Stop")
        snippet = """
import Path
import PathScripts
from PathScripts import PathUtils
prjexists = False
obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython","Stop")
Path.Op.Gui.Stop.Stop(obj)

Path.Op.Gui.Stop._ViewProviderStop(obj.ViewObject)
PathUtils.addToJob(obj)
"""
        FreeCADGui.doCommand(snippet)
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_Stop", CommandPathStop())


FreeCAD.Console.PrintLog("Loading PathStop… done\n")



def create_nodes(femmesh):
    # nodes
    femmesh.addNode(0.0, 2000.0, 0.0, 1)
    femmesh.addNode(0.0, 0.0, 0.0, 2)
    femmesh.addNode(2000.0, 2000.0, 0.0, 3)
    femmesh.addNode(2000.0, 0.0, 0.0, 4)
    femmesh.addNode(4000.0, 2000.0, 0.0, 5)
    femmesh.addNode(4000.0, 0.0, 0.0, 6)
    femmesh.addNode(6000.0, 2000.0, 0.0, 7)
    femmesh.addNode(6000.0, 0.0, 0.0, 8)
    femmesh.addNode(8000.0, 2000.0, 0.0, 9)
    femmesh.addNode(8000.0, 0.0, 0.0, 10)
    femmesh.addNode(10000.0, 2000.0, 0.0, 11)
    femmesh.addNode(10000.0, 0.0, 0.0, 12)
    femmesh.addNode(12000.0, 2000.0, 0.0, 13)
    femmesh.addNode(12000.0, 0.0, 0.0, 14)
    femmesh.addNode(1000.0, 1000.0, 2000.0, 15)
    femmesh.addNode(3000.0, 1000.0, 2000.0, 16)
    femmesh.addNode(5000.0, 1000.0, 2000.0, 17)
    femmesh.addNode(7000.0, 1000.0, 2000.0, 18)
    femmesh.addNode(9000.0, 1000.0, 2000.0, 19)
    femmesh.addNode(11000.0, 1000.0, 2000.0, 20)
    return True


def create_elements(femmesh):
    # elements
    femmesh.addEdge([1, 2], 1)
    femmesh.addEdge([3, 4], 2)
    femmesh.addEdge([5, 6], 3)
    femmesh.addEdge([7, 8], 4)
    femmesh.addEdge([9, 10], 5)
    femmesh.addEdge([11, 12], 6)
    femmesh.addEdge([13, 14], 7)
    femmesh.addEdge([1, 3], 8)
    femmesh.addEdge([2, 4], 9)
    femmesh.addEdge([3, 5], 10)
    femmesh.addEdge([4, 6], 11)
    femmesh.addEdge([5, 7], 12)
    femmesh.addEdge([6, 8], 13)
    femmesh.addEdge([7, 9], 14)
    femmesh.addEdge([8, 10], 15)
    femmesh.addEdge([9, 11], 16)
    femmesh.addEdge([10, 12], 17)
    femmesh.addEdge([11, 13], 18)
    femmesh.addEdge([12, 14], 19)
    femmesh.addEdge([1, 15], 20)
    femmesh.addEdge([3, 16], 21)
    femmesh.addEdge([5, 17], 22)
    femmesh.addEdge([7, 18], 23)
    femmesh.addEdge([9, 19], 24)
    femmesh.addEdge([11, 20], 25)
    femmesh.addEdge([2, 15], 26)
    femmesh.addEdge([4, 16], 27)
    femmesh.addEdge([6, 17], 28)
    femmesh.addEdge([8, 18], 29)
    femmesh.addEdge([10, 19], 30)
    femmesh.addEdge([12, 20], 31)
    femmesh.addEdge([15, 3], 32)
    femmesh.addEdge([16, 5], 33)
    femmesh.addEdge([17, 7], 34)
    femmesh.addEdge([18, 9], 35)
    femmesh.addEdge([19, 11], 36)
    femmesh.addEdge([20, 13], 37)
    femmesh.addEdge([15, 4], 38)
    femmesh.addEdge([16, 6], 39)
    femmesh.addEdge([17, 8], 40)
    femmesh.addEdge([18, 10], 41)
    femmesh.addEdge([19, 12], 42)
    femmesh.addEdge([20, 14], 43)
    femmesh.addEdge([15, 16], 44)
    femmesh.addEdge([16, 17], 45)
    femmesh.addEdge([17, 18], 46)
    femmesh.addEdge([18, 19], 47)
    femmesh.addEdge([19, 20], 48)
    femmesh.addEdge([2, 3], 49)
    femmesh.addEdge([4, 5], 50)
    femmesh.addEdge([6, 7], 51)
    femmesh.addEdge([8, 9], 52)
    femmesh.addEdge([10, 11], 53)
    femmesh.addEdge([12, 13], 54)
    return True
